/*
 * File Name   : vdce_init.c
 * 
 * Description :
 *                This file containts functions for initializing VDCE communication 
 *                systems to transfer data between application modules.
 *
 * Functions : vdce_init()
 *             recvfromcvm()
 *             setup_datapath()
 *
 * Output(s) : open socket port and get socket ports
 *
 *
 * Programmed by : Yoonhee Kim, IlKyun Ra
 *                 Syracuse University
 *
 * Date : 15 December, 1996
 *
 */

#include 	<string.h>
#include	"inet.h"
#include 	"dvm.h"
#include 	"dvm_msg.h"
#include	<sys/errno.h>

#define		FileType	(0)
#define		CONNECT_TRY	120

extern		int	errno;
extern		struct task *_dvminfo;
extern		int	_total_data;
extern		int	_myvdcetaskid;
extern		int	_myvdcetaskname;
extern		FILE	*jfp;
extern		int 	_myuserid;
extern		int 	_myapplid;
extern		int 	_cvmsockfd;
extern		int	_mypid;

extern		VDCE_MSG_HEADER     _cvm_header;

extern		ACT_format *_actinfo;


/*
 * Function Name : vdce_init
 *
 * Description   : Setup an execution enviroment for a user application.
 *
 * Input(s)      : 
 *		   cvmarc - no. of arguments associated with CVM
 *		   cvmargv - list of arguments associated with CVM
 *                 argc - no. of arguments for the tasks
 *                 argv - argument list of tasks
 *
 * Output(s)     : initialize task execution information and setup data
 *                 path.
 */

int  rc;	/* return result 1 : success
			         0 : error
                */

vdce_init(cvmargc,cvmargv,argc,argv)
int             cvmargc;
char            **cvmargv;
int             *argc;
char            **argv;
{

    int                 trymax;
    int   	        cvmsockport;
    int 		i,n,j;
    int                 mytaskid, no_of_channels;
    int 		cvmsockfd;
    struct sockaddr_in	cvm_addr;	/* data strcuture for cvm socket */
    extern	char 	*gethostaddr();
    char		msg[2048],temp[100], *recv_msg;
    VDCE_MSG_HEADER     msg_header;
    int		   	myuserid, myapplid, faultflag;	
    char		command[5], msg_data[2048];
    


     trymax = CONNECT_TRY;

    /*
     * get execution environment data from application CVM manager
     */


     if (cvmargc < 4) {
#ifdef DEBUG
	fprintf(jfp,"vdce_init: Please give me more input data\n");
        fprintf(jfp,"%s:socketport taskid total_no_of_channels",argv[0]);
#endif
/*taskname -r #ofchannel -1 userid applid taskid master -p4pg progroupfilename*/
	rc = 0;
        return(rc);
     }

     _total_data = no_of_channels  = atoi(cvmargv[2]);
     cvmsockport = atoi(cvmargv[3]); /* -1 or socketportnumber*/
     _myuserid=myuserid    = atoi(cvmargv[4]);
     _myapplid=myapplid    = atoi(cvmargv[5]);
     _myvdcetaskid=mytaskid    = atoi(cvmargv[6]);
#ifdef DEBUG
     fprintf(jfp,"no_of_channels  = %d\n",_total_data);
    fflush(jfp); 
     fprintf(jfp,"socket port = %d\n",cvmsockport);
    fflush(jfp); 
     fprintf(jfp,"my task id  = %d\n",mytaskid);
    fflush(jfp); 
#endif

    if (cvmsockport != STDINOUT) {
       /* using socket for communicating cvm */
       /*
       * Fill in the structure "cvm_addr" with the address of the
       * CVM that DVM wants to connect with
       */

       bzero((char *) &cvm_addr, sizeof(cvm_addr));
       cvm_addr.sin_family	= AF_INET;
       cvm_addr.sin_addr.s_addr	= inet_addr(gethostaddr());
       cvm_addr.sin_port		= htons(cvmsockport);


      /* 
       * Open a TCP socket for communicating CVM
       */

      while (trymax>0) {
      if ( (cvmsockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	 err_sys("VDCE_INIT: DVM can't open stream socket for CVM");
	 exit(1);
	}
#ifdef DEBUG
     fprintf(jfp,"after open : fd = %d\n", cvmsockfd);
     fflush(jfp); 
#endif
      /*
       * Connet to the CVM
       */
            if (connect(cvmsockfd, (struct sockaddr *) &cvm_addr,
			   sizeof(cvm_addr)) < 0)  {
	    fprintf(jfp,"+%d", errno);
     	    fflush(jfp); 
            trymax--; /*try  100times/seconds*/
	    close(cvmsockfd);
	    sleep(1);
            }
	    else {
		 trymax = -1;
		 break ;
	    }

#ifdef DEBUG
     fprintf(jfp,"after connect command : fd = %d\n", cvmsockfd);
     fflush(jfp); 
#endif
      }

      if (trymax==0) {
	  err_ret(
          "vdce_init: I am sorry I can't connect CVM. \n Please check CVM\n");
	  return(0);
      }
      _cvmsockfd = cvmsockfd;
    }
    else 
	cvmsockfd = STDINOUT;

#ifdef DEBUG
     fprintf(jfp,"success connect  : fd = %d\n", cvmsockfd);
     fflush(jfp); 
#endif
    /*
     * allocation memory for task information structure
     */
    if ( (_dvminfo = (struct task *) 
          malloc(sizeof(struct task)*no_of_channels)) ==NULL) 
        	err_dump("vdce_init: No memory");

    /*
     * memory allocation for act feature structure
     */
    if ( (_actinfo = (ACT_format *)
	  malloc(sizeof(ACT_format))) == NULL)
		err_dump("vdce_init : No memory");


    /*
     * Read execution environment information from CVM
     */

    /*
     * starts getting task execution environment information from CVM 
    */
/*CIT info */

     faultflag = recvfromcvmCIT(cvmsockfd,_dvminfo,mytaskid,no_of_channels,argc,argv);
     if (faultflag == 0) { 
		err_ret("vdce_init: error from recvfromcvmCIT\n");
	 	return (0);	
      }

/*send back CIT Ack */

 if (cvmsockport != STDINOUT) {
/* create cvm_header for management communication */
      strcpy(_cvm_header.H_name, HEADER_MESSAGE);
      strcpy(_cvm_header.message_type,DVM2LCVM);
      sprintf(_cvm_header.destination_addr, "%s;%d",gethostaddr(),
	cvmsockport);
      bzero(_cvm_header.source_addr, DUMMYSIZE);
      create_header_address(&_cvm_header.source_addr[DUMMYSIZE], myuserid, 
	myapplid, mytaskid);

/*CIT Ack message , There is the number of lines after
command, in this case 3 lines. */
      sprintf(msg, "%s\n%d\n%d\n%d\n%d\n%d\n", CIT_Conf_ACK, 4, myuserid, 
	myapplid, mytaskid,_mypid); 
			/*userid, applid, taskid, _mypid*/
/* add _mypid into CIT Ack Message to support user task abortion  1/16/98 */

      sprintf(_cvm_header.length, "%d",strlen(msg)+HEADERSIZE); 

#ifdef DEBUG
	fprintf(jfp,"my processor id = %d\n",_mypid);
	fprintf(jfp, "send msg = %s", msg);
	fflush(jfp);
#endif
      send_data_using_msg(cvmsockfd, _cvm_header,msg, strlen(msg));

/* wait for Task_Activation message or ... */

      recv_data_using_msg(cvmsockfd, &msg_header, &recv_msg);
#ifdef DEBUG
        fprintf(jfp," received %s\n ",recv_msg);
	fflush(jfp);
#endif
      
      decode_command(recv_msg, command, msg_data);

      if (strncmp(command, Task_Activation, 4)) {
		return (0); 
	}


      /* ACT feature */

/*
      recv_data_using_msg(cvmsockfd, &msg_header, &recv_msg);
#ifdef DEBUG
        fprintf(jfp," received %s\n ",recv_msg);
#endif
 
      decode_command(recv_msg, command, msg_data);
*/
 
     
     }


/*ACT table is ready */


	/*Can continue applicaiontion path set up */
    /* 
     * Insert arguments related to application
    if ((_actinfo->comm_type == P4) || (_actinfo->comm_type == MPI)) 
    */

    	n = *argc;
    	for(i=8;i<cvmargc;i++) {
       		argv[n] = (char *) malloc(MAXPATHLEN*sizeof(char));
       		sprintf(argv[n],"%s",cvmargv[i]);
       		n++;
    	}
    	*argc = n;

    /*
     * Estasblish all data channels related to my task 
    */

#ifdef DEBUG
    for(i=1;i<*argc;i++) 
	fprintf(jfp,"argv[%d]=%s\n",i,argv[i]);
	fflush(jfp);

#endif
    setup_datapath(_dvminfo,mytaskid,no_of_channels);

#ifdef DEBUG
	fprintf(jfp,"vdce_init done\n");
	fflush(jfp);
#endif


    return (1);
} /* end vdce_init */


/*
 * Function Name : recvfromcvmCIT()
 *
 * Description   : 
 *     DVM receives execution environment information data from CVM via      
 *     the sepecified socket port.
 *
 * Input(s)      : sockfd - cvm socket port descriptor
 *                 tkchinfo - task channel information
 *                 mytaskid - my task id
 *                 no_of_channels - no of data associated the submitted application.
 *                 argc - no. of arguments for the tasks
 *                 argv - argument list of tasks
 *
 * Output(s)     : fill up the exection environment struct 
 *                 using received data from CVM.
 *
 */
int recvfromcvmCIT(sockfd,tkchinfo,mytaskid,no_of_channels,argc,argv)
int sockfd;
struct	task	*tkchinfo;
int     mytaskid,no_of_channels;
int     *argc;
char    *argv[];
{

    int                inch[500];
    int                 pinch, poutch;
    int                outch[500];

    int i,j,tmpid,n, tmplength;
    char *x;

    VDCE_MSG_HEADER *vdce_msg;
    int msgsize;
    char *msgbuf, *token, temp[100];

    n=1;

    /*
     * get the task environment execution information
    */
#ifdef DEBUG
     fprintf(jfp,"sockfd in recvfromcvmCIT= %d\n",sockfd);
     fflush(jfp);
#endif


    vdce_msg = (VDCE_MSG_HEADER *)malloc(sizeof(VDCE_MSG_HEADER));

    pinch=poutch=0;
    for (i=0;i<no_of_channels;i++) {

       if (sockfd != STDINOUT) {
        /*
         * Data read from CVM socket port
        */
#ifdef DEBUG		
     fprintf(jfp,"before read header in recvfromcvmCIT= %d\n",sockfd);
     fflush(jfp);
#endif
        bzero(vdce_msg, HEADERSIZE);
        if ( read(sockfd,vdce_msg,HEADERSIZE) != 
             HEADERSIZE) {
#ifdef DEBUG
	fprintf(jfp,"vdce_init header:%s\n%s\n%s\n%s\n%s\n", 
	vdce_msg->H_name,vdce_msg->message_type, 
vdce_msg->destination_addr, vdce_msg->source_addr, vdce_msg->length);
     fflush(jfp);
#endif
	err_ret("vdce_init header: Fail to receive CVM data from CVM\n");
	return(0);
        }
	
        strncpy(temp, vdce_msg->length,8);
        msgsize = atoi(temp); 
	msgsize -= HEADERSIZE;
	
#ifdef DEBUG
	fprintf(jfp,"vdce_init header:%s\n%s\n%s\n%s\n%s\n", 
	vdce_msg->H_name,vdce_msg->message_type, 
vdce_msg->destination_addr, vdce_msg->source_addr, vdce_msg->length);

	fprintf (jfp,"msgsize = %d(%s)\n", msgsize, temp);
     fflush(jfp);

#endif
        msgbuf = (char *) malloc((msgsize+8)*sizeof(char));

        if ( (tmplength = read(sockfd,msgbuf,msgsize)) < 
             msgsize) {
#ifdef DEBUG
        	fprintf(jfp, "vdce_init size = %d msg = %s\n", tmplength,msgbuf);
		for (i = 0 ; i < msgsize ; i++)
			fprintf(jfp, "%c", msgbuf[i]);
		 fflush(jfp);
#endif
		err_ret("vdce_init msg: Fail to receive CVM data from CVM");
		return(0);
        } 
#ifdef DEBUG
        fprintf(jfp, "vdce_init msg = %s\nprint each char", msgbuf);
	fflush(jfp);
#endif
        token = strtok(msgbuf, "\n");
        if (strncmp(token, CIT_Conf, 4)) {/*it is not CIT message */
			err_ret( "vdce_init :Wrong CIT information");
                        return(0);
        }
        token = strtok(0, "\n");
        tkchinfo[i].user_id = atoi(token);
        token = strtok(0, "\n");
        tkchinfo[i].appl_id = atoi(token);

        token = strtok(0, "\n");
        strcpy(tkchinfo[i].v_name,  token);
        token = strtok(0, "\n");
        tkchinfo[i].v_type = atoi(token);
        token = strtok(0, "\n");
        strcpy(tkchinfo[i].s_name, token);
        token = strtok(0, "\n");
        tkchinfo[i].s_pid = atoi(token);
        token = strtok(0, "\n");
        tkchinfo[i].s_port = atoi(token);
        token = strtok(0, "\n");
        tkchinfo[i].s_tkid = atoi(token);
        token = strtok(0, "\n");
        tkchinfo[i].s_socket = atoi(token);
/* data representation */
        token = strtok(0, "\n");
        tkchinfo[i].s_data_rep = atoi(token);

        token = strtok(0, "\n");
        strcpy(tkchinfo[i].d_name, token);
        token = strtok(0, "\n");
        tkchinfo[i].d_pid = atoi(token);
        token = strtok(0, "\n");
        tkchinfo[i].d_port = atoi(token);
        token = strtok(0, "\n");
        tkchinfo[i].d_tkid = atoi(token);
        token = strtok(0, "\n");
        tkchinfo[i].d_socket = atoi(token);
/*data representation */
        token = strtok(0, "\n");
        tkchinfo[i].d_data_rep = atoi(token);
	
      }
      else {
       /*
        * Data read from jfp 
       */

        fscanf(stdin,"%s",&tkchinfo[i].v_name[0]);
	fscanf(stdin,"%d",&tkchinfo[i].v_type);
	fscanf(stdin,"%s",&tkchinfo[i].s_name[0]);
	fscanf(stdin,"%d",&tkchinfo[i].s_pid);
	fscanf(stdin,"%d",&tkchinfo[i].s_port);
	fscanf(stdin,"%d",&tkchinfo[i].s_tkid);
	fscanf(stdin,"%d",&tkchinfo[i].s_socket);
	fscanf(stdin,"%s",&tkchinfo[i].d_name[0]);
	fscanf(stdin,"%d",&tkchinfo[i].d_pid);
	fscanf(stdin,"%d",&tkchinfo[i].d_port);
	fscanf(stdin,"%d",&tkchinfo[i].d_tkid);
	fscanf(stdin,"%d",&tkchinfo[i].d_socket);
      } /* if end */

#ifdef DEBUG
        fprintf(jfp,"print struct task\n%d ",tkchinfo[i].user_id);
	fflush(jfp); 
        fprintf(jfp,"%d ",tkchinfo[i].appl_id);
	fflush(jfp); 
        fprintf(jfp,"%s ",tkchinfo[i].v_name);
	fflush(jfp); 
        fprintf(jfp,"%d ",tkchinfo[i].v_type);
	fflush(jfp); 
        fprintf(jfp,"%s ",tkchinfo[i].s_name);
	fflush(jfp); 
        fprintf(jfp,"%d ",tkchinfo[i].s_pid);
	fflush(jfp); 
        fprintf(jfp,"%d ",tkchinfo[i].s_port);
	fflush(jfp); 
        fprintf(jfp,"%d ",tkchinfo[i].s_tkid);
	fflush(jfp); 
        fprintf(jfp,"%d ",tkchinfo[i].s_socket);
	fflush(jfp); 
        fprintf(jfp,"%d ",tkchinfo[i].s_data_rep);
	fflush(jfp); 
        fprintf(jfp,"%s ",tkchinfo[i].d_name);
	fflush(jfp); 
        fprintf(jfp,"%d ",tkchinfo[i].d_pid);
	fflush(jfp); 
        fprintf(jfp,"%d ",tkchinfo[i].d_port);
	fflush(jfp); 
        fprintf(jfp,"%d ",tkchinfo[i].d_tkid);
	fflush(jfp); 
        fprintf(jfp,"%d\n",tkchinfo[i].d_socket);
	fflush(jfp); 
        fprintf(jfp,"%d ",tkchinfo[i].d_data_rep);
	fflush(jfp); 
#endif


    
        /*
         * set arguments for tasks
        */
	if (mytaskid == tkchinfo[i].d_tkid) {
                /* check inputs */
		inch[pinch]=tkchinfo[i].d_pid; /* save input pids*/
		pinch++;
	} /* end if */

    } /* end for */


    /*
     * Check outputs
    */
    for (i=0;i<no_of_channels;i++) {
        if (tkchinfo[i].s_tkid==mytaskid) {
		outch[poutch] = tkchinfo[i].s_pid; /* save output pids */
                poutch++;
        }
    }

    /* sorting for input channel id */
    for(i=0;i<pinch;i++) {
	for(j=i;j<pinch;j++) {
	  if (inch[i] > inch[j]) {
		tmpid = inch[i];
		inch[i] = inch[j];
		inch[j] = tmpid;
	  }
	}
    }

   
    /* sorting for output channel id */
    for(i=0;i<poutch;i++) {
	for(j=i;j<poutch;j++) {
	  if (outch[i] > outch[j]) {
		tmpid = outch[i];
		outch[i] = outch[j];
		outch[j] = tmpid;
	  }
	}
    }

#ifdef DEBUG
   fprintf(jfp,"vdce_int:SORTING IS DONE\n");
   fflush(jfp);
#endif

    /* store input channel ids to argv */
    n = 1; i=0;
    while (i < pinch) {
	argv[n] = (char *) malloc(MAXPATHLEN*sizeof(char));
        sprintf(argv[n],"%d",inch[i]);
	n++; i++;
    }
#ifdef DEBUG
   fprintf(jfp,"vdce_int:Save Inputchannels DONE\n");
   fflush(jfp);
#endif
    /* store input channel ids to argv */
    i=0;
    while (i < poutch) {
	argv[n] = (char *) malloc(MAXPATHLEN*sizeof(char));
        sprintf(argv[n],"%d",outch[i]);
	n++; i++;
    }
    

#ifdef DEBUG
    for(i=1;i<n;i++)
                fprintf(jfp,"vdce_init: argv[%d] = %s\n",i,argv[i]);
   fflush(jfp);
#endif
    
    *argc = n; /* set no. of arguments */
return (1);
} /* end recvfromcvm */

/*
 * Function Name : setup_datapath()
 *
 * Description   :
 *                 Setup all data path associated with my task 
 *
 * Input(s)      : tkchinfo - execution environment information
 *                 mytaskid - my task id 
 *                 no_of_channels - no. of data 
 *
 * Ouput(s)      : socket descriptors    
 */
setup_datapath(tkchinfo,mytaskid,no_of_channels)
struct	task	*tkchinfo;
int	mytaskid;
int	no_of_channels;
{
  int i;

  /*
   * check my task id with source taskid and destination taskid
   */
  for(i=0;i<no_of_channels;i++) {
     /*
      * Check sending data port 
      */
    if (tkchinfo[i].s_pid != -1 && tkchinfo[i].d_pid != -1 &&
	tkchinfo[i].s_tkid == mytaskid && tkchinfo[i].v_type!=0) {
        /*
         * Preparing sending socket descriptor and
         * setup connection
        */
        setup_sending_path(&tkchinfo[i]);
   fflush(jfp);
    }

    /*
     * Check receiving port 
    */
    if (tkchinfo[i].s_pid != -1 && tkchinfo[i].d_pid!=-1 &&
	tkchinfo[i].d_tkid == mytaskid && tkchinfo[i].v_type!=0) {
	/*
         * preparing receiving socket descriptor
         * and setup connection
         */
         setup_recv_path(&tkchinfo[i]);
   fflush(jfp);
    }


/*
#ifdef DEBUG
	fprintf(jfp,"Variable name         = %s\n",tkchinfo[i].v_name);
	fprintf(jfp,"Variable type         = %d\n",tkchinfo[i].v_type);
	fprintf(jfp,"source host name      = %s\n",tkchinfo[i].s_name);
	fprintf(jfp,"source task id        = %d\n",tkchinfo[i].s_tkid);
	fprintf(jfp,"source port id        = %d\n",tkchinfo[i].s_pid);
	fprintf(jfp,"source sock port      = %d\n",tkchinfo[i].s_port);
	fprintf(jfp,"source socket descrip = %d\n",tkchinfo[i].s_socket);
	fprintf(jfp,"destination host name = %s\n",tkchinfo[i].d_name);
	fprintf(jfp,"destination port id   = %d\n",tkchinfo[i].d_pid);
	fprintf(jfp,"destination taskid    = %d\n",tkchinfo[i].d_tkid);
	fprintf(jfp,"destination sock port = %d\n",tkchinfo[i].d_port);
	fprintf(jfp,"destination socket de = %d\n",tkchinfo[i].d_socket);
	fflush(jfp);
#endif
*/
  } /* end for */

} /* end setup_datapath */

/*
 * Function Name : setup_sending_path()
 *
 * Description   : open a socket for sending data to a task and
 *                 make a connection via socket. The socket  
 *                 descriptor is saved at the task execution 
 *                 environment information.
 *
 * Input(s)      : tkchinfo - its task execution environment 
 *                            information
 *
 * Output(s)     : opened socket descriptor is saved.
 *
 */
setup_sending_path(tkchinfo)
struct task	*tkchinfo;
{

    char *msg;
    int sockfd,newsockfd;
    int time_out = 0;
    int is_bind  = 0;

    struct sockaddr_in    mytask_addr,hetask_addr;
    int    hetasklen;

    /*
     * Open a TCP socket (an Internet stream socket).
    */

    if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
          err_ret("VDCE_INIT: can't open stream socket");
	  return;
    }

    /*
     * Bind our local address so that the client can send to us.
    */

    bzero((char *) &mytask_addr, sizeof(mytask_addr));
    mytask_addr.sin_family      = AF_INET;
/*
    mytask_addr.sin_addr.s_addr = inet_addr(&tkchinfo->d_name[0]);
*/
    mytask_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    mytask_addr.sin_port        = htons(tkchinfo->d_port);


#ifdef DEBUG
    fprintf(jfp,"VDCE_INIT: (%d) bind socket %d with %s \n",tkchinfo->d_tkid,
            tkchinfo->d_port,&tkchinfo->d_name[0]);
#endif

    while (time_out <= CONNECT_TRY) {
     /*
      * Try for about CONNECT_TRY sec
     */
     if (bind(sockfd,(struct sockaddr *) &mytask_addr, sizeof(mytask_addr)) <0) {
         time_out++;
         sleep(1);
     }
     else
         break; /* success to bind */
    }

    /*
     * Check binding status
    */
     if (time_out > CONNECT_TRY) {
        msg = (char *) calloc(128,sizeof(char));
#ifdef DEBUG
        fprintf(jfp,"VDCE_INIT: %d task can't bind with a task %d\n",
                tkchinfo->s_tkid,tkchinfo->d_tkid);

#endif
        return;
     }

#ifdef DEBUG
     fprintf(jfp,"** Bind OK\n");
#endif

     /*
      * listen
     */

     listen(sockfd, 5);

     time_out = 0;

     /*
      * Try for CONNECT_TRY sec
     */

     hetasklen = sizeof(hetask_addr);
     while (time_out <= CONNECT_TRY) {
        /*
         * Wait for a connection from a client process.
        */

        if ( (newsockfd = accept(sockfd,(struct sockaddr *) &hetask_addr,
                                 &hetasklen)) < 0) {
#ifdef DEBUG
	     fprintf(jfp,"vdce_init: accept return value = %d\n",newsockfd);
#endif
             time_out++;
             sleep(1);
        }
        else
              break; /* socket connection setup */
     }

     if (time_out > CONNECT_TRY) {
        msg = (char *) calloc(128,sizeof(char));
        sprintf(msg,"VDCE_INIT: task %d can't connect with task %d\n",
                tkchinfo->s_tkid,tkchinfo->d_tkid);
        close(sockfd);
        err_sys(msg);
     }
     else 
        /*
         * Save socket port for future sending - connection is fine!
        */
#ifdef DEBUG
       fprintf(jfp,"newsock descriptor = %d\n",newsockfd);
#endif  
        tkchinfo->s_socket = newsockfd;
  
} /* end setup_sending_path */

/*
 * Function Name : setup_recv_path()
 *
 * Description   : open a socket for receiving data to a task and
 *                 make a connection via socket. The socket  
 *                 descriptor is saved at the task execution 
 *                 environment information.
 *
 * Input(s)      : tkchinfo - its task execution environment 
 *                            information
 *
 * Output(s)     : opened socket descriptor is saved.
 *
 */
setup_recv_path(tkchinfo)
struct task     *tkchinfo;
{
  
  int portnum, sockfd;
  struct sockaddr_in src_addr;
  int optval = 1;

  int is_connect = 0;
  int time_out   = 0;

  /*
   * Open a TCP socket to obtain data from parent task
   */
  bzero((char *) &src_addr,sizeof(src_addr));
  src_addr.sin_family       = AF_INET;
  src_addr.sin_addr.s_addr  = inet_addr(&tkchinfo->s_name[0]);
  src_addr.sin_port         = htons(tkchinfo->s_port);

/*
  setsockopt(sockfd,SOL_SOCKET,SO_REUSEADDR,(char *)&optval,sizeof(optval));
*/

 time_out  = 0; is_connect = 0;
  while ( ! is_connect && (time_out < CONNECT_TRY)) { /* connect try uptime CONNECT_TRY times */
    if ( (sockfd = socket(AF_INET,SOCK_STREAM, 0)) < 0) {
        err_ret(
        "VDCE_RECV: can't open a socket to receive data from a task");
	return;
    }
       /*
        * Make connection to other task
        */

#ifdef DEBUG
fprintf(jfp,"VDCE_INIT: Task (%d) is trying to connect to %s via socket(%d).\n",
        tkchinfo->d_tkid,&tkchinfo->s_name[0],tkchinfo->s_port);
	fflush(jfp);
#endif

        if ( connect(sockfd,(struct sockaddr *) &src_addr,
                     sizeof(src_addr)) < 0) {
	    fprintf(jfp,"data connect errno =%d\n",errno);
	fflush(jfp);
            time_out ++;
	    sleep(1);
	    close(sockfd);
        }
        else {
             is_connect = 1; /* success to connect */
	     break;
        }
  } /* while end */


  /*
   * Receive data from former task if socket is established.
  */
  if (is_connect && (time_out < CONNECT_TRY)) {
        /* Save socket descriptor */
	tkchinfo->d_socket = sockfd; 
#ifdef DEBUG
    fprintf(jfp,"VDCE_INIT: Task (%d) connect socket %d to Task(%d) %s \n",
    tkchinfo->d_tkid, tkchinfo->s_port,tkchinfo->s_tkid,&tkchinfo->s_name[0]);
    fprintf(jfp,"VDCE_INIT: new socket file descriptor is saved\n");
	fflush(jfp);
#endif
	return;
  }
  else  {
    fprintf(jfp,"VDCE_INIT: Task (%d) couldn't connet socket %d to Task(%d) %s \n",
    tkchinfo->d_tkid, tkchinfo->s_port,tkchinfo->s_tkid,&tkchinfo->s_name[0]);
	fprintf(jfp,"VDCE_INIT: can't setup receiving socket\n"); 
	fflush(jfp);
	vdce_error(125);
	return;
  }
} /* end setup_recv_path */

