/*
 * File Name   : vdce_init.c
 * 
 * Description :
 *                This file containts functions for initializing VDCE communication 
 *                systems to transfer data between application modules.
 *
 * Functions : vdce_init()
 *             recvfromcvm()
 *             setup_datapath()
 *
 * Output(s) : open socket port and get socket ports
 *
 *
 * Programmed by : IlKyun Ra
 *                 Syracuse University
 *
 * Date : 15 December, 1996
 *
 */

#include	"inet.h"
#include 	"dvm.h"
#include	<sys/errno.h>

extern		int	errno;
extern		struct task *_dvminfo;
extern		int	_total_data;
/*
 * Function Name : vdce_init
 *
 * Description   : Setup an execution enviroment for a user application.
 *
 * Input(s)      : 
 *		   cvmarc - no. of arguments associated with CVM
 *		   cvmargv - list of arguments associated with CVM
 *                 argc - no. of arguments for the tasks
 *                 argv - argument list of tasks
 *
 * Output(s)     : initialize task execution information and setup data
 *                 path.
 */

int  rc;	/* return result 1 : success
			         0 : error
                */

vdce_init(cvmargc,cvmargv,argc,argv)
int             cvmargc;
char            **cvmargv;
int             *argc;
char            **argv;
{

    int                 trymax=50;
    int   	        cvmsockport;
    int 		i;
    int                 mytaskid, no_of_data;
    int 		cvmsockfd;
    struct sockaddr_in	cvm_addr;	/* data strcuture for cvm socket */
    extern	char 	*gethostaddr();


    /*
     * get execution environment data from application CVM manager
     */

     if (cvmargc < 4) {
	fprintf(stderr,"vdce_init: Please give me more input data\n");
        fprintf(stderr,"%s:socketport taskid total_no_of_data",argv[0]);
	rc = 0;
        return(rc);
     }

     cvmsockport = atoi(cvmargv[1]);
     mytaskid    = atoi(cvmargv[2]);
     _total_data = no_of_data  = atoi(cvmargv[3]);
     fprintf(stderr,"socket port = %d\n",cvmsockport);
     fprintf(stderr,"my task id  = %d\n",mytaskid);
     fprintf(stderr,"no_of_data  = %d\n",_total_data);

    /*
     * allocation memory for task information structure
     */
    if ( (_dvminfo = (struct task *) malloc(sizeof(struct task)*no_of_data)) ==NULL)
        	err_dump("vdce_init: No memory");

    /*
     * Fill in the structure "cvm_addr" with the address of the
     * CVM that DVM wants to connect with
     */

     bzero((char *) &cvm_addr, sizeof(cvm_addr));
     cvm_addr.sin_family	= AF_INET;
     cvm_addr.sin_addr.s_addr	= inet_addr(gethostaddr());
     cvm_addr.sin_port		= htons(cvmsockport);


    /* 
     * Open a TCP socket for communicating CVM
     */

    if ( (cvmsockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) 
	 err_sys("VDCE_INIT: DVM can't open stream socket for CVM");

    /*
     * Connet to the CVM
     */

    while (trymax>0) {
          if (connect(cvmsockfd, (struct sockaddr *) &cvm_addr,
			   sizeof(cvm_addr)) < 0)  {
	  fprintf(stderr,"+");
          trymax--;
	  sleep(1);
          }
	  else break;
    }

    if (trymax<=0) {
	fprintf(stderr,
        "vdce_init: I am sorry I can't connet CVM. \n Please check CVM\n");
	return(0);
    }
 
    /*
     * Read execution environment information from CVM
     */


    /*
     * starts getting task execution environment information from CVM 
    */
    recvfromcvm(cvmsockfd,_dvminfo,mytaskid,no_of_data,argc,argv);         

    close(cvmsockfd); /* close cvm socket port */

    /*
     * Estasblish all data channels related to my task 
    */

    setup_datapath(_dvminfo,mytaskid,no_of_data);

    return;
} /* end vdce_init */

/*
 * Function Name : recvfromcvm()
 *
 * Description   : 
 *     DVM receives execution environment information data from CVM via      
 *     the sepecified socket port.
 *
 * Input(s)      : sockfd - cvm socket port descriptor
 *                 taskinfo - task execution information
 *                 mytaskid - my task id
 *                 no_of_data - no of data associated the submitted application.
 *                 argc - no. of arguments for the tasks
 *                 argv - argument list of tasks
 *
 * Output(s)     : fill up the exection environment struct 
 *                 using received data from CVM.
 *
 */
recvfromcvm(sockfd,taskinfo,mytaskid,no_of_data,argc,argv)
int sockfd;
struct	task	*taskinfo;
int     mytaskid,no_of_data;
int     *argc;
char    *argv[];
{

    int i,n;


    n=1;

    /*
     * get the task environment execution information
    */
    for (i=0;i<no_of_data;i++) {

        /*
         * Data read from CVM socket port
        */
        if (read(sockfd,&taskinfo[i],sizeof(struct task)) != sizeof(struct task)) 
		err_ret("vdce_init: Fail to receive CVM data from CVM");
          
        /*
         * set arguments for tasks
        */
	if (mytaskid == taskinfo[i].d_tkid && taskinfo[i].d_pid != -1) {

                /* check inputs */
		argv[n] = (char *) calloc(MAXPATHLEN,sizeof(char));
		sprintf(argv[n],"%d",taskinfo[i].d_pid);
/***
		strcpy(argv[n],taskinfo[i].v_name);
***/
#ifdef DEBUG
		fprintf(stderr,"vdce_init: argv[%d] = %s\n",n,argv[n]);
#endif
		n++;
	} /* end if */

    } /* enf for */

    /*
     * Check outputs
    */
    for (i=0;i<no_of_data;i++) {
        if (taskinfo[i].s_tkid==mytaskid && taskinfo[i].s_pid != -1) {
                argv[n] = (char *) calloc(MAXPATHLEN,sizeof(char));
		sprintf(argv[n],"%d",taskinfo[i].s_pid);
/***
                strcpy(argv[n],taskinfo[i].v_name);
***/
#ifdef DEBUG
                fprintf(stderr,"vdce_init: argv[%d] = %s\n",n,argv[n]);
#endif
                n++;
        }
        else if (taskinfo[i].d_pid == -1 && taskinfo[i].s_tkid == mytaskid) {
                argv[n] = (char *) calloc(MAXPATHLEN,sizeof(char));
                strcpy(argv[n],taskinfo[i].v_name);
#ifdef DEBUG
                fprintf(stderr,"vdce_init: argv[%d] = %s\n",n,argv[n]);
#endif
                n++;

        }
    }

    *argc = n; /* set no. of arguments */

} /* end recvfromcvm */

/*
 * Function Name : setup_datapath()
 *
 * Description   :
 *                 Setup all data path associated with my task 
 *
 * Input(s)      : taskinfo - execution environment information
 *                 mytaskid - my task id 
 *                 no_of_data - no. of data 
 *
 * Ouput(s)      : socket descriptors    
 */
setup_datapath(taskinfo,mytaskid,no_of_data)
struct	task	*taskinfo;
int	mytaskid;
int	no_of_data;
{
  int i;

  /*
   * check my task id with source taskid and destination taskid
   */
  for(i=0;i<no_of_data;i++) {
     /*
      * Check sending data port 
      */
    if (taskinfo[i].s_pid != -1 && taskinfo[i].d_pid != -1 &&
	taskinfo[i].s_tkid == mytaskid) {
        /*
         * Preparing sending socket descriptor and
         * setup connection
        */
        setup_sending_path(&taskinfo[i]);
    }

    /*
     * Check receiving port 
    */
    if (taskinfo[i].s_pid != -1 && taskinfo[i].d_pid!=-1 &&
	taskinfo[i].d_tkid == mytaskid) {
	/*
         * preparing receiving socket descriptor
         * and setup connection
         */
         setup_recv_path(&taskinfo[i]);
    }

#ifdef DEBUG
/*
	fprintf(stdout,"Variable name         = %s\n",taskinfo[i].v_name);
	fprintf(stdout,"Variable type         = %d\n",taskinfo[i].v_type);
	fprintf(stdout,"source host name      = %s\n",taskinfo[i].s_name);
	fprintf(stdout,"source task id        = %d\n",taskinfo[i].s_tkid);
	fprintf(stdout,"source port id        = %d\n",taskinfo[i].s_pid);
	fprintf(stdout,"source sock port      = %d\n",taskinfo[i].s_port);
	fprintf(stdout,"source socket descrip = %d\n",taskinfo[i].s_socket);
	fprintf(stdout,"destination host name = %s\n",taskinfo[i].d_name);
	fprintf(stdout,"destination port id   = %d\n",taskinfo[i].d_pid);
	fprintf(stdout,"destination taskid    = %d\n",taskinfo[i].d_tkid);
	fprintf(stdout,"destination sock port = %d\n",taskinfo[i].d_port);
	fprintf(stdout,"destination socket de = %d\n",taskinfo[i].d_socket);
	fflush(stdout);
*/
#endif
  } /* end for */

} /* end setup_datapath */

/*
 * Function Name : setup_sending_path()
 *
 * Description   : open a socket for sending data to a task and
 *                 make a connection via socket. The socket  
 *                 descriptor is saved at the task execution 
 *                 environment information.
 *
 * Input(s)      : taskinfo - its task execution environment 
 *                            information
 *
 * Output(s)     : opened socket descriptor is saved.
 *
 */
setup_sending_path(taskinfo)
struct task	*taskinfo;
{

    char *msg;
    int sockfd,newsockfd;
    int time_out = 0;
    int is_bind  = 0;

    struct sockaddr_in    mytask_addr,hetask_addr;
    int    hetasklen;

    /*
     * Open a TCP socket (an Internet stream socket).
    */

    if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
          fprintf(stderr,"VDCE_INIT: can't open stream socket");
	  return;
    }

    /*
     * Bind our local address so that the client can send to us.
    */

    bzero((char *) &mytask_addr, sizeof(mytask_addr));
    mytask_addr.sin_family      = AF_INET;
/*
    mytask_addr.sin_addr.s_addr = inet_addr(&taskinfo->d_name[0]);
*/
    mytask_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    mytask_addr.sin_port        = htons(taskinfo->d_port);


    fprintf(stderr,"VDCE_INIT: (%d) bind socket %d with %s \n",taskinfo->d_tkid,
            taskinfo->d_port,&taskinfo->d_name[0]);

    while (time_out <= 50) {
     /*
      * Try for about 50 sec
     */
     if (bind(sockfd,(struct sockaddr *) &mytask_addr, sizeof(mytask_addr)) <0) {
         time_out++;
         sleep(1);
     }
     else
         break; /* success to bind */
    }

    /*
     * Check binding status
    */
     if (time_out > 50) {
        msg = (char *) calloc(128,sizeof(char));
        fprintf(stderr,"VDCE_SEND: %d task can't bind with a task %d\n",
                taskinfo->s_tkid,taskinfo->d_tkid);
        return;
     }

#ifdef DEBUG
     fprintf(stderr,"** Bind OK\n");
#endif

     /*
      * listen
     */

     listen(sockfd, 5);

     time_out = 0;

     /*
      * Try for 50 sec
     */

     hetasklen = sizeof(hetask_addr);
     while (time_out <= 50) {
        /*
         * Wait for a connection from a client process.
        */

        if ( (newsockfd = accept(sockfd,(struct sockaddr *) &hetask_addr,
                                 &hetasklen)) < 0) {
	     fprintf(stderr,"vdce_init: accept return value = %d\n",newsockfd);
             time_out++;
             sleep(1);
        }
        else
              break; /* socket connection setup */
     }

     if (time_out > 50) {
        msg = (char *) calloc(128,sizeof(char));
        sprintf(msg,"VDCE_SEND: task %d can't connect with task %d\n",
                taskinfo->s_tkid,taskinfo->d_tkid);
        close(sockfd);
        err_sys(msg);
     }
     else 
        /*
         * Save socket port for future sending - connection is fine!
        */
#ifdef DEBUG
       fprintf(stderr,"newsock descriptor = %d\n",newsockfd);
#endif  
        taskinfo->s_socket = newsockfd;
  
} /* end setup_sending_path */

/*
 * Function Name : setup_recv_path()
 *
 * Description   : open a socket for receiving data to a task and
 *                 make a connection via socket. The socket  
 *                 descriptor is saved at the task execution 
 *                 environment information.
 *
 * Input(s)      : taskinfo - its task execution environment 
 *                            information
 *
 * Output(s)     : opened socket descriptor is saved.
 *
 */
setup_recv_path(taskinfo)
struct task     *taskinfo;
{
  
  int portnum, sockfd;
  struct sockaddr_in src_addr;
  int optval = 1;

  int is_connect = 0;
  int time_out   = 0;

  /*
   * Open a TCP socket to obtain data from parent task
   */
  bzero((char *) &src_addr,sizeof(src_addr));
  src_addr.sin_family       = AF_INET;
  src_addr.sin_addr.s_addr  = inet_addr(&taskinfo->s_name[0]);
  src_addr.sin_port         = htons(taskinfo->s_port);

/*
  setsockopt(sockfd,SOL_SOCKET,SO_REUSEADDR,(char *)&optval,sizeof(optval));
*/

    if ( (sockfd = socket(AF_INET,SOCK_STREAM, 0)) < 0) {
        fprintf(stderr,
        "VDCE_RECV: can't open a socket to receive data from a task");
	return;
    }
       /*
        * Make connection to other task
        */

#ifdef DEBUG
fprintf(stderr,"VDCE_INIT: Task (%d) is trying to connect to %s via socket(%d).\n",
        taskinfo->d_tkid,&taskinfo->s_name[0],taskinfo->s_port);
#endif
 time_out  = 0; is_connect = 0;

  while ( ! is_connect && time_out < 50) { /* connect try uptime 50 times */
        if ( connect(sockfd,(struct sockaddr *) &src_addr,
                     sizeof(src_addr)) < 0) {
	    fprintf(stderr,"errno =%d\n",errno);
            time_out ++;
	    sleep(1);
        }
        else {
             is_connect = 1; /* success to connect */
	     break;
        }
  } /* while end */

#ifdef DEBUG
    fprintf(stderr,"VDCE_INIT: Task (%d) connet socket %d to Task(%d) %s \n",
    taskinfo->d_tkid, taskinfo->s_port,taskinfo->s_tkid,&taskinfo->s_name[0]);
#endif

  /*
   * Receive data from former task if socket is established.
  */
  if (is_connect && time_out < 50) {
        /* Save socket descriptor */
	taskinfo->d_socket = sockfd; 
	fprintf(stderr,"VDCE_INIT: new socket file descriptor is saved\n");
	return;
  }
  else  {
	fprintf(stderr,"VDCE: can't setup receiving socket\n"); 
	return;
  }
} /* end setup_recv_path */

