/**************************************************************************   
 *                                                                        *
 *    File Name : inv.c                                                   *
 *                                                                        *
 *    Author    : KIM, DONGMIN (kimd@top.cis.syr.edu)                     *
 *                                                                        *
 *    Function  : This program is a master program for computing inverse  *
 *                matrix for VDCE application.                            *
 *                This program read data from buffer which stores         *
 *                informations received by vdce_recv.                     *
 *                After the computation, it stores inverse matrix to      *
 *                buffer so that they can be sent by vdce_send.           *
 *                                                                        *
 **************************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <sys/types.h>
#include <unistd.h>
#include <math.h>
#include <string.h>

#include "p4.h"

#include "vdce.h"

#define DATA 1

#define inter_comm "??"

extern	char	*vdce_get_my_name();
extern	char	*vdce_input_name();
extern	char	*vdce_output_name();
extern	int	vdce_data_represent();

vdce_run(argc,argv)
     int  argc;
     char *argv[];
{
  int nsize;
  int inch1, outch1;
  char	**buf;
  char tmp[10];
  char *temp;
  int i, j, k;
  int matrix_size, type, size;
  int nslave, my_id, to, row, load, from, column;
  int LU;
  float **Matrix, **I;
  float *incoming;
  float L, U, divide, mult;
  float element, zero;
  char	*out1_msg;

  long  total_elapsed;
  FILE  *fd, *fd1;
  char  *task_name;
  int   task_id;
  char  file_name[100];
  char  *io_file;
  int   *counter;
  char  *buf2;
  int   *tmp_matrix_size;
  int   total_size;
  char  *send_buf;
  float *tmp_float;

  timer_reset();

  task_name = (char *) malloc(10*sizeof(char ));
  if (task_name == NULL) {
    fprintf(fd, "malloc error at task_name");
    fflush(fd);
  }

  task_name = vdce_get_my_name();
  task_id = vdce_get_my_id();

#ifdef DEMOLOG
  sprintf(file_name, "/merlin2/users/demo/vdce/ver1/log/inv_master%d.inter", task_id);
  fd = fopen(file_name, "w");
#endif

#ifdef DEBUG
  sprintf(file_name, "/tmp/inv%d.inter", task_id);
  fd = fopen(file_name, "w");
#endif

#ifdef DEBUG
  fprintf(fd, "before p4_initenv\n");
  fflush(fd);
#endif /*DEBUG */
  p4_initenv(&argc, argv);

#ifdef DEBUG
  fprintf(fd, "after p4_initenv\n");
  fflush(fd);
#endif
  my_id = p4_get_my_id();
  if (my_id != 0) {
    slave(); 
    p4_wait_for_end();
    exit(1); 
  }          

  p4_create_procgroup();
  nslave = p4_num_total_ids() - 1;
#ifdef DEBUG
  fprintf(fd, "after p4_create_procgroup\n");
  fprintf(fd,"my_id :%d  total proc : %d\n", my_id, nslave);
  fflush(fd);
#endif

  inch1 = atoi(argv[1]);
  outch1 = atoi(argv[2]);

#ifdef DEBUG
  fprintf(fd,"in_port1 = %d, out_port1 = %d\n",
	  inch1,outch1);
  fflush(fd);
#endif

#ifdef DEBUG
  fprintf(fd,"task id = %d\n", task_id);
  fflush(fd);
#endif

  io_file = (char *) malloc(100*sizeof(char ));
  bzero(io_file, 100);

  io_file = vdce_input_name(task_id, inch1);
  if (io_file == NULL) {
    fprintf(fd, "malloc error at io_file");
    fflush(fd);
  }
#ifdef DEBUG
  fprintf(fd,"io file name : %s\n", io_file); 
  fflush(fd);
#endif

#ifdef DEMOLOG
  fprintf(fd,"io file name : %s\n", io_file); 
  fflush(fd);
#endif

  if ( !strcmp(io_file, inter_comm) ) {
    /* inter process communication */
    /* VDCE receive */
#ifdef DEBUG
    fprintf(fd,"vdce receive\n"); 
    fflush(fd);
#endif

    nsize=vdce_recv(inch1, &buf2);

    if ( vdce_data_represent(inch1, INPUT) ) { /* different data type */

      temp = strtok((char * ) buf2, " ");
      matrix_size = atoi(temp);
      Matrix =  (float **) malloc(matrix_size * sizeof(float *));
      if (Matrix == NULL) {
	fprintf(fd, "malloc error at Matrix");
      }
      for ( i = 0 ; i < matrix_size ; i++ ) {
	Matrix[i] = (float *) malloc(matrix_size * sizeof(float));
	if (Matrix[i] == NULL) {
	  fprintf(fd, "malloc error at Matrix[%d]", i);
	}
      }
      for ( i = 1 ; i <= matrix_size ; i++ ) {
	temp = strtok( 0, " ");
	Matrix[i-1][0] = (float)atof(temp);
	for ( j = 1 ; j < matrix_size ; j++ ) {
	  temp = strtok( 0, " ");
	  Matrix[i-1][j] = (float)atof(temp);
	}
      }
    }
    else { /* same data type */

      tmp_matrix_size = (int *)buf2;

      matrix_size = *tmp_matrix_size;

#ifdef DEMOLOG
      fprintf(fd, "receivec size : %d   Matrix size : %d\n", nsize, matrix_size);
      fflush(fd);
#endif

      Matrix =  (float **) malloc(matrix_size * sizeof(float *));
      if (Matrix == NULL) {
	fprintf(fd, "malloc error at Matrix");
      }

      for ( i = 0 ; i < matrix_size ; i++ ) {
	Matrix[i] = (float *)(buf2 + sizeof(int) + i * sizeof(float) * matrix_size);
      }

#ifdef DEBUG
      fprintf(fd, "Original Matrix : \n");
      for ( i = 0 ; i < matrix_size ; i++ ) {
	for ( j = 0 ; j < matrix_size ; j++ ) {
	  fprintf(fd, "%f ", Matrix[i][j]);
	  fflush(fd);
	}
	fprintf(fd, "\n");
      }

      fprintf(fd, "receivd matrix size :%d", matrix_size);
      fflush(fd);
#endif

      if (nsize <= 0 ) {
#ifdef DEBUG
	fprintf(fd,"data has not been arrived\n"); 
	fflush(fd);
#endif
	exit(1);
      }
    }
  }
  else {
    /* external file */
    /* get input data from file */
#ifdef DEBUG
    fprintf(fd,"external file\n"); 
    fflush(fd);
#endif
    fd1 = fopen(io_file, "r");
 
    if ( fd1 == NULL ) {
      fprintf(fd, "\ncannot find file!\n");
      fflush(fd);
      exit(1);
    }
 
    /* read the size of the matrix. */
    fscanf( fd1, "%d", &matrix_size );

    Matrix =  (float **) malloc(matrix_size * sizeof(float *));
    if (Matrix == NULL) {
      fprintf(fd, "malloc error at Matrix");
    }
    for ( i = 0 ; i < matrix_size ; i++ ) {
      Matrix[i] = (float *) malloc(matrix_size * sizeof(float));
      if (Matrix[i] == NULL) {
        fprintf(fd, "malloc error at Matrix[%d]", i);
      }
    }

    /* read from the file */
    for ( i = 0 ; i < matrix_size ; i++ ) {
      for ( j = 0 ; j < matrix_size ; j++ ) {
	fscanf(fd1, "%f", &element);
	Matrix[i][j] = element;
      }
    }
    fclose(fd1);
  }

  load = matrix_size / nslave;

#ifdef DEBUG
  fprintf(fd, "load : %d\n", load);
  fflush(fd);
#endif

#ifdef DEMOLOG
  fprintf(fd, "load : %d\n", load);
  fflush(fd);
#endif

  I =  (float **) malloc(matrix_size * sizeof(float *));
  if (I == NULL) {
    fprintf(fd, "malloc error at I");
  }

  for ( i = 0 ; i < matrix_size ; i++ ) {
    I[i] = (float *) malloc(matrix_size * sizeof(float));
    if (I[i] == NULL) {
      fprintf(fd, "malloc error at I[%d]",i);
    }
  }

  /* Initiate the Identity matrix */
  for ( i = 0 ; i < matrix_size ; i++ ) {
    for ( j = 0 ; j < matrix_size ; j++ ) {
      if ( i == j ) {
	I[i][j] = 1.0;
      }
      else {
	I[i][j] = 0.0;
      }
    }
  }

/**/
#ifdef RECVDEBUG
  fprintf(fd, "\nOriginal Matrix\n");
  for ( i = 0 ; i < matrix_size ; i++ ) {
    for ( j = 0 ; j < matrix_size ; j++ ) {
      fprintf(fd, "%5.3f ", Matrix[i][j]);
  fflush(fd);
    }
    fprintf(fd, "\n");
  }
  fflush(fd);
#endif

  L = 0.;
  U = 0.;

  /* check if the matrix is L, U or neither */
  for ( i = 0 ; i < matrix_size ; i++ ) {
    for ( j = 0 ; j < matrix_size ; j++ ) {
      if ( i < j ) {
	L = L + Matrix[i][j];
      }
      else if ( i > j ) {
	U = U + Matrix[i][j];
      }
      else {
      }
    }
  }

  if ( L == 0.0 ) { /* the matrix is lower matrix */
    LU = 1;
  }
  else if ( U == 0.0 ) {  /* the matrix is upper matrix */
    LU = 2;
  }
  else { /* neither */
    LU = 0;
  }

    /* convert the matrix to upper matrix */
  if ( LU == 0 ) {
    LU == 2;
    for ( i = 0 ; i < matrix_size - 1 ; i++ ) {
      for ( j = i + 1 ; j < matrix_size ; j++ ) {
	divide = Matrix[j][i];
	mult = Matrix[i][i];
	if ( divide != 0 ) {
	  for ( k = 0 ; k < matrix_size ; k++ ) {
/**
	     printf("   before    k = %d   Matrix[%d][%d] = %f    Matrix[%d][%d] = %f\n", k, j, k, Matrix[j][k], i, k, Matrix[i][k]); 
**/
	    Matrix[j][k] = Matrix[i][k] - ((mult / divide) * Matrix[j][k]);
/**
	     printf("   after    k = %d   Matrix[%d][%d] = %f    Matrix[%d][%d] = %f\n", k, j, k, Matrix[j][k], i, k, Matrix[i][k]); 
**/
	    I[k][j] = I[k][i] - ((mult / divide) * I[k][j]);
	  }
	}
	Matrix[j][i] = 0.;
      }
    }
  }

#ifdef RECVDEBUG
  fprintf(fd, "Matrix\n");
  for ( i = 0 ; i < matrix_size ; i++ ) {
    for ( j = 0 ; j < matrix_size ; j++ ) {
      fprintf(fd, "%5.3f ", Matrix[i][j]);
    }
    fprintf(fd, "\n");
  }
  fflush(fd);

  fprintf(fd, "I Matrix\n");
  for ( i = 0 ; i < matrix_size ; i++ ) {
    for ( j = 0 ; j < matrix_size ; j++ ) {
      fprintf(fd, "%5.3f ", I[j][i]);
    }
    fprintf(fd, "\n");
  }
  fflush(fd);

#endif
  
#ifdef DEMOLOG
  fprintf(fd, "convert done\n");
  fflush(fd);
#endif

/* distribute data */
  for ( to = 1 ; to <= nslave ; to++ ) {
    p4_send(DATA, to, &matrix_size, sizeof(int));
    p4_send(DATA, to, &load, sizeof(int));
    p4_send(DATA, to, &LU, sizeof(int));

    for ( i = 0 ; i < matrix_size ; i++ ) {
      p4_send(DATA, to, Matrix[i], matrix_size * sizeof(float));
    }

    for ( i = 0 ; i < load ; i++ ) {
      p4_send(DATA, to, I[(to - 1) * load + i], matrix_size * sizeof(float));
    }

  }
/**
  p4_broadcast(DATA, &matrix_size, sizeof(int));
  p4_broadcast(DATA, &load, sizeof(int));
  p4_broadcast(DATA, &LU, sizeof(int));
  for ( row = 0 ; row < matrix_size ; row++ ) {
    p4_broadcast(DATA, Matrix[row], matrix_size * sizeof(float));
  }

  for ( to = 1 ; to <= nslave ; to++ ) {
    for ( row = 0 ; row < load ; row++ ) {
      p4_send(DATA, to,  I[(to - 1) * load + row], matrix_size * sizeof(float));
    }
  }

**/

#ifdef DEBUG
  fprintf(fd, "INV distribute is done\n");
  fflush(fd);
#endif

#ifdef DEMOLOG
  fprintf(fd, "INV distribute is done\n");
  fflush(fd);
#endif

  /* collect the result */
/**
  size	= matrix_size * sizeof(float);
  type    = DATA;

  for ( i = 0 ; i < load ; i++) {
    for ( from = 1 ; from <= nslave ; from++ ) {
      column = (from - 1) * load;

      incoming	= NULL;
      p4_recv(&type, &from, &incoming, &size);
      for ( j = 0 ; j < matrix_size ; j++ ) {
	Matrix[j][column + i]	= incoming[j];
      }
      p4_msg_free(incoming);
    }
  }
**/

  counter = (int *) malloc(nslave * sizeof(int));
  for ( i = 0 ; i < nslave ; i++ ) {
    counter[i] = 0;
  }

  size	= matrix_size * sizeof(float);
  type    = DATA;

  for ( i = 0 ; i < matrix_size ; i++) {
    from = -1;
    incoming = NULL;
    p4_recv(&type, &from, &incoming, &size);
    column = (from - 1) * load;

    for ( j = 0 ; j < matrix_size ; j++ ) {
      Matrix[j][column + counter[from - 1]] = incoming[j];
    }
    p4_msg_free(incoming);
    counter[from-1] = counter[from-1] + 1;
  }

#ifdef DEBUG
  for ( i = 0 ; i < nslave ; i++ ) {
    fprintf(fd, "%d : %d\n", i+1, counter[i]);
  }
  fflush(fd);
#endif

  free(I);

  p4_wait_for_end();

/*
  printf("\nResult Matrix\n");
  for ( i = 0 ; i < matrix_size ; i++ ) {
    for ( j = 0 ; j < matrix_size ; j++ ) {
      printf("%5.3f ", Matrix[i][j]);
    }
    printf("\n");
  }

*/
#ifdef SENDDEBUG
  fprintf(fd, "\nResult Matrix\n");
  for ( i = 0 ; i < matrix_size ; i++ ) {
    for ( j = 0 ; j < matrix_size ; j++ ) {
      fprintf(fd, "%8.3f ", Matrix[i][j]);
    }
    fprintf(fd, "\n");
  }
  fflush(fd);
#endif

#ifdef DEBUG
    total_elapsed = timer_elapsed();
    fprintf(fd, "After Receive : %s%d  %d\n", task_name, task_id, total_elapsed);
    fflush(fd);
#endif

  bzero(io_file, 100);
  io_file = vdce_output_name(task_id, outch1);
#ifdef DEBUG
  fprintf(fd,"io file name : %s\n", io_file); 
  fflush(fd);
#endif

  if ( !strcmp(io_file, inter_comm) ) {
    /* inter process communication */
    /* VDCE receive */

    if ( vdce_data_represent(outch1, OUTPUT) ) { /* different data type */

      /* write to the buffer */
      send_buf = (char *) malloc((matrix_size * matrix_size * 10 + 10) * sizeof(char));
      if (send_buf == NULL) {
        fprintf(fd, "malloc error at send_buf");
      }
      total_size = sizeof(send_buf);
      bzero(tmp, 10);
      bzero(send_buf, (matrix_size * matrix_size * 10 + 10));
      sprintf(tmp, "%d ", matrix_size);
      strcat(send_buf, tmp);
      element = 1.0;
      zero = 0.0;
      for ( i = 0 ; i < matrix_size ; i++ ) {
	for ( j = 0 ; j < matrix_size ; j++ ) {
	  bzero(tmp, 10);
	  sprintf( tmp, "%5.3f ", Matrix[i][j]);
	  strcat(send_buf, tmp);
	}
      }
      strcat(send_buf,"\n");
    }
    else {

      total_size = sizeof(matrix_size) + sizeof(float) * matrix_size * matrix_size;

#ifdef DEMOLOG
      fprintf(fd, "total size : %d\n", total_size);
      fflush(fd);
#endif

      send_buf = (char *)malloc(total_size + 1);

      bcopy(&matrix_size, send_buf, sizeof(matrix_size));

#ifdef DEMOLOG
      fprintf(fd, "check 1\n");
      fflush(fd);
#endif

      for ( i = 0 ; i < matrix_size ; i++ ) {
	bcopy(Matrix[i], send_buf + sizeof(matrix_size) + i * sizeof(float) * matrix_size, sizeof(float) * matrix_size);
      }

#ifdef DEMOLOG
      fprintf(fd, "check 2\n");
      fflush(fd);
#endif

    }
    vdce_send(outch1, send_buf, total_size);
  }
  else {
    /* external file */
    fd1 = fopen(io_file, "w");
 
    if ( fd1 == NULL ) {
      fprintf(fd, "\ncannot find file!\n");
    }
    else {
      fprintf(fd1, "%d\n", matrix_size);
      for ( i = 0 ; i < matrix_size ; i++ ) {
	for ( j = 0 ; j < matrix_size ; j++ ) {
	  fprintf( fd1, "%7.3f", Matrix[i][j]);
	}
	fprintf(fd1, "\n");
      }
      fclose(fd1);
    }
  }

#ifdef DEMOLOG
  fprintf(fd, "All is done\n");
  fflush(fd);
  fclose(fd);
#endif

#ifdef DEBUG
  fprintf(fd, "All is done\n");
  fflush(fd);
  fclose(fd);
#endif


#ifdef DEBUG
#ifdef DEMOLOG
  fd = fopen("/merlin2/users/demo/vdce/ver1/log/time.dat", "a+");
#else
  fd = fopen("/tmp/time.dat", "a+");
#endif
#endif

 total_elapsed = timer_elapsed();

#ifdef DEBUG
  if (fd!=NULL) {
    fprintf(fd, "%s%d  %d\n", task_name, task_id, total_elapsed);
    fflush(fd);
    fclose(fd);  
  }
#endif

/*send CVM execution time */
 report_execution_time(2,total_elapsed);

}

vdce_def(mytaskinfo)
     VDCE_TASKINFO	*mytaskinfo;
{

  vdce_set_taskname(mytaskinfo,"INV");
	
  vdce_set_inputs(mytaskinfo,"L");

  vdce_set_outputs(mytaskinfo,"INV_L");
}
