/**************************************************************************   
 *                                                                        *
 *    File Name : lu.c                                                    *
 *                                                                        *
 *    Author    : KIM, DONGMIN (kimd@top.cis.syr.edu)                     *
 *                                                                        *
 *    Function  : This program is a master program for LU decomposition   *
 *                for VDCE application.                                   *
 *                This program read data from buffer which stores         *
 *                informations received by vdce_recv.                     *
 *                After the computation, it stores L and U matrix to      *
 *                buffers so that they can be sent by vdce_send.          *
 *                                                                        *
 **************************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <sys/types.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include <strings.h>

#include "p4.h"

#include "vdce.h"

#define DATA 1
#ifdef DEMOLOG
#define time_file "/merlin2/users/demo/vdce/ver1/log/time.dat"
#else
#define time_file "/tmp/time.dat"
#endif

#define inter_comm "??"

extern	char	*vdce_get_my_name();
extern	char	*vdce_input_name();
extern	char	*vdce_output_name();
extern	int	vdce_data_represent();

vdce_run(argc,argv)
     int  argc;
     char *argv[];
{
  int nsize;
  int inch1, outch1, outch2;
  char	**buf;
  char tmp[10];
  char *temp;
  int i, j, k;
  int matrix_size, type, size;
  int nslave, my_id, to, row, load, from;
  float **Matrix, **submatrix, **L, **U;

  float *incoming;
  float element, zero;
  char	*out1_msg;
  char	*out2_msg;
  char  file_name[100];

  long  total_elapsed;
  FILE  *fd, *fd1;
  char  *task_name;
  int   task_id;
  char  *io_file;
  int   *counter;
  char  *buf2;
  int   *tmp_matrix_size;
  int   total_size;
  char  *send_buf1;
  char  *send_buf2;

  timer_reset();

  task_name = (char *) malloc(10*sizeof(char ));
  if (task_name == NULL) {
    fprintf(fd, "malloc error at task_name");
    fflush(fd);
  }

  task_name = vdce_get_my_name();
  task_id = vdce_get_my_id();

#ifdef DEMOLOG
  sprintf(file_name, "/merlin2/users/demo/vdce/ver1/log/lu%d.inter", task_id);
#else
  sprintf(file_name, "/tmp/lu%d.inter", task_id);
#endif

  fd = fopen(file_name, "w");

#ifdef DEBUG
  fprintf(fd, "in vdce_run 1\n");
  fflush(fd);
#endif

  p4_initenv(&argc, argv);
#ifdef DEBUG
  fprintf(fd, "in vdce_run 2\n");
  fflush(fd);
#endif
  my_id = p4_get_my_id();
  if(my_id != 0){
    slave(); 
    p4_wait_for_end();
#ifdef DEBUG
  fprintf(fd, "in vdce_run3 \n");
  fflush(fd);
#endif
    exit(0); 
  }          
#ifdef DEBUG
  fprintf(fd, "in vdce_run 4\n");
  fflush(fd);
#endif
  p4_create_procgroup();
#ifdef DEBUG
  fprintf(fd, "in vdce_run 5\n");
  fflush(fd);
#endif

  nslave = p4_num_total_ids() - 1;
#ifdef DEBUG
  fprintf(fd, "my_id :%d  total proc : %d\n", my_id, nslave);
  fflush(fd);
#endif

  inch1 = atoi(argv[1]);
  outch1 = atoi(argv[2]);
  outch2 = atoi(argv[3]);

#ifdef DEBUG
  fprintf(fd,"in_port1 = %d, out_port1 = %d , out_port2 = %d\n",
	  inch1,outch1,outch2);
  fflush(fd);
#endif

#ifdef DEBUG
  fprintf(fd,"task id = %d\n", task_id);
  fflush(fd);
#endif

  io_file = (char *) malloc(100*sizeof(char ));
  bzero(io_file, 100);

  io_file = vdce_input_name(task_id, inch1);
  if (io_file == NULL) {
    fprintf(fd, "malloc error at io_file");
    fflush(fd);
  }
#ifdef DEBUG
  fprintf(fd,"io file name : %s\n", io_file); 
  fflush(fd);
#endif

  if ( !strcmp(io_file, inter_comm) ) {
    /* inter process communication */
    /* VDCE receive */
#ifdef DEBUG
    fprintf(fd,"vdce receive\n"); 
    fflush(fd);
#endif

    nsize=vdce_recv(inch1, &buf2);

    if ( vdce_data_represent(inch1, INPUT) ) { /* different data type */

      temp = strtok((char * ) buf2, " ");
      matrix_size = atoi(temp);
      Matrix =  (float **) malloc(matrix_size * sizeof(float *));
      if (Matrix == NULL) {
	fprintf(fd, "malloc error at Matrix");
      }
      for ( i = 0 ; i < matrix_size ; i++ ) {
	Matrix[i] = (float *) malloc(matrix_size * sizeof(float));
	if (Matrix[i] == NULL) {
	  fprintf(fd, "malloc error at Matrix[%d]", i);
	}
      }
      for ( i = 1 ; i <= matrix_size ; i++ ) {
	temp = strtok( 0, " ");
	Matrix[i-1][0] = (float)atof(temp);
	for ( j = 1 ; j < matrix_size ; j++ ) {
	  temp = strtok( 0, " ");
	  Matrix[i-1][j] = (float)atof(temp);
	}
      }
    }
    else { /* same data type */

      tmp_matrix_size = (int *)buf2;

      matrix_size = *tmp_matrix_size;

      Matrix =  (float **) malloc(matrix_size * sizeof(float *));
      if (Matrix == NULL) {
	fprintf(fd, "malloc error at Matrix");
      }

      for ( i = 0 ; i < matrix_size ; i++ ) {
	Matrix[i] = (float *)(buf2 + sizeof(int) + i * sizeof(float) * matrix_size);
      }

      if (nsize <= 0 ) {
#ifdef DEBUG
	fprintf(fd,"data has not been arrived\n"); 
	fflush(fd);
#endif
	exit(1);
      }
    }
  }
  else {
    /* external file */
    /* get input data from file */
#ifdef DEBUG
    fprintf(fd,"external file\n"); 
    fflush(fd);
#endif
    fd1 = fopen(io_file, "r");
 
    if ( fd1 == NULL ) {
      printf("\ncannot find file!\n");
      exit(1);
    }
 
    /* read the size of the matrix. */
    fscanf( fd1, "%d", &matrix_size );
    Matrix =  (float **) malloc(matrix_size * sizeof(float *));
    if (Matrix == NULL) {
      fprintf(fd, "malloc error at Matrix");
    }
    for ( i = 0 ; i < matrix_size ; i++ ) {
      Matrix[i] = (float *) malloc(matrix_size * sizeof(float));
      if (Matrix[i] == NULL) {
        fprintf(fd, "malloc error at Matrix[%d]", i);
      }
    }

    /* read from the file */
    for ( i = 0 ; i < matrix_size ; i++ ) {
      for ( j = 0 ; j < matrix_size ; j++ ) {
	fscanf(fd1, "%f", &element);
	Matrix[i][j] = element;
      }
    }
    fclose(fd1);
  }

  load = matrix_size / nslave;

#ifdef DEBUG
  for ( i = 0 ; i < matrix_size ; i++ ) {
    for ( j = 0 ; j < matrix_size ; j++ ) {
      fprintf(fd, "%f  ", Matrix[i][j]);
    }
    fprintf(fd, "\n");
  }
  fflush(fd);
#endif

  free(buf);
  /* distribute data */
  for ( to = 1 ; to <= nslave ; to++ ) {
    p4_send(DATA, to, &load, sizeof(int));
    p4_send(DATA, to, &matrix_size, sizeof(int));

    for ( i = 0 ; i < load ; i++ ) {
      row = nslave * i + (to - 1);
      p4_send(DATA, to, Matrix[row], matrix_size * sizeof(float));
    }
  }

#ifdef DEBUG
  fprintf(fd,"DATA distribution is done\n");
  fflush(fd);
#endif

  /* collect the result */
/**
  size	= matrix_size * sizeof(float);
  type    = DATA;
  for ( from = 1 ; from <= nslave ; from++ ) {
    row = from - 1;

    for ( i = 0 ; i < load ; i++) {

      incoming	= NULL;
      p4_recv(&type, &from, &incoming, &size);
      for ( j = 0 ; j < matrix_size ; j++ ) {
	Matrix[row][j]	= incoming[j];
	Matrix[row][j]	= Matrix[row][j];
      }
      row = row + nslave;
      p4_msg_free(incoming);
    }
  }
**/

  L =  (float **) malloc(matrix_size * sizeof(float *));
  if (L == NULL) {
    fprintf(fd, "malloc error at L");
  }
  for ( i = 0 ; i < matrix_size ; i++ ) {
    L[i] = (float *) malloc(matrix_size * sizeof(float));
    if (L[i] == NULL) {
      fprintf(fd, "malloc error at L[%d]", i);
    }
  }

  U =  (float **) malloc(matrix_size * sizeof(float *));
  if (U == NULL) {
    fprintf(fd, "malloc error at U");
  }
  for ( i = 0 ; i < matrix_size ; i++ ) {
    U[i] = (float *) malloc(matrix_size * sizeof(float));
    if (U[i] == NULL) {
      fprintf(fd, "malloc error at U[%d]", i);
    }
  }

  counter = (int *) malloc(nslave * sizeof(int));
  for ( i = 0 ; i < nslave ; i++ ) {
    counter[i] = 0;
  }

  size	= matrix_size * sizeof(float);
  type    = DATA;

  for ( i = 0 ; i < matrix_size ; i++) {
    /* receive the data */
    /* receive the data by row */
    from = -1;
    incoming	= NULL;
    p4_recv(&type, &from, &incoming, &size);
    row = counter[from - 1] * nslave + (from - 1);
    for ( j = 0 ; j < matrix_size ; j++ ) {
      if ( row > j ) { /* lower matrix */
	L[row][j] = incoming[j];
	U[row][j] = 0.;
      }
      else if ( row < j ) { /* upper matrix */
	L[row][j] = 0.;
	U[row][j] = incoming[j];
      }
      else { /* row == j */
	L[row][j] = 1.;
	U[row][j] = incoming[j];
      }
    }
    p4_msg_free(incoming);
    counter[from-1] = counter[from-1] + 1;
  }

  p4_wait_for_end();

#ifdef SENDDEBUG
  fprintf(fd,"\nResult Matrix\n");
  for ( i = 0 ; i < matrix_size ; i++ ) {
    for ( j = 0 ; j < matrix_size ; j++ ) {
      fprintf(fd,"%5.3f ", Matrix[i][j]);
    }
    fprintf(fd,"\n");
  }
  fflush(fd);
#endif

#ifdef DEBUG
  fprintf(fd,"\nResult Matrix\n");
  fflush(fd);
#endif

  bzero(io_file, 100);
  io_file = vdce_output_name(task_id, outch1);
#ifdef DEBUG
  fprintf(fd,"io file name : %s\n", io_file); 
  fflush(fd);
#endif

  if ( !strcmp(io_file, inter_comm) ) {
    /* inter process communication */
    /* VDCE receive */
#ifdef DEBUG
    fprintf(fd,"vdce send L\n"); 
    fflush(fd);
#endif

    if ( vdce_data_represent(outch1, OUTPUT) ) { /* different data type */

      /* write to the buffer */
      /* write L part */
      send_buf1 = (char *) malloc((matrix_size * matrix_size * 10 + 10) * sizeof(char));
      if (send_buf1 == NULL) {
        fprintf(fd, "malloc error at send_buf1");
      }
      total_size = sizeof(send_buf1);
      bzero(tmp, 10);
      bzero(send_buf1, (matrix_size * matrix_size * 10 + 10));
      sprintf(tmp, "%d ", matrix_size);
      strcat(send_buf1, tmp);
      element = 1.0;
      zero = 0.0;
      for ( i = 0 ; i < matrix_size ; i++ ) {
	for ( j = 0 ; j < matrix_size ; j++ ) {
	  bzero(tmp, 10);
	  sprintf( tmp, "%5.3f ", L[i][j]);
	  strcat(send_buf1, tmp);
	}
      }
      strcat(send_buf1,"\n");

#ifdef DEBUG
      fprintf(fd, "L :\n \n" );
#endif

    }
    else {

      total_size = sizeof(matrix_size) + sizeof(float) * matrix_size * matrix_size;

#ifdef DEMOLOG
      fprintf(fd, "total size : %d\n", total_size);
      fflush(fd);
#endif
      send_buf1 = (char *)malloc(total_size + 1);

      bcopy(&matrix_size, send_buf1, sizeof(matrix_size));

#ifdef DEMOLOG
      fprintf(fd, "check 1\n");
      fflush(fd);
#endif
      for ( i = 0 ; i < matrix_size ; i++ ) {
	bcopy(L[i], send_buf1 + sizeof(matrix_size) + i * sizeof(float) * matrix_size, sizeof(float) * matrix_size);
      }

#ifdef DEMOLOG
      fprintf(fd, "check 2\n");
      fflush(fd);
#endif

#ifdef DEBUG
      fprintf(fd,"\nL Matrix\n");
      for ( i = 0 ; i < matrix_size ; i++ ) {
	for ( j = 0 ; j < matrix_size ; j++ ) {
	  fprintf(fd,"%5.3f ", L[i][j]);
	}
	fprintf(fd,"\n");
      }
      fflush(fd);
#endif
    }

    vdce_send(outch1, send_buf1, total_size);
  }
  else {
    /* external file */
    fd1 = fopen(io_file, "w");
 
    if ( fd1 == NULL ) {
      fprintf(fd, "\ncannot find file!\n");
    }
    else {
      fprintf(fd1, "%d\n", matrix_size);
      for ( i = 0 ; i < matrix_size ; i++ ) {
	for ( j = 0 ; j < matrix_size ; j++ ) {
	  fprintf( fd1, "%7.3f", L[i][j]);
	}
	fprintf(fd1, "\n");
      }
      fclose(fd1);
    }
  }

  bzero(io_file, 100);
  io_file = vdce_output_name(task_id, outch2);
#ifdef DEBUG
  fprintf(fd,"io file name : %s\n", io_file); 
  fflush(fd);
#endif

  if ( !strcmp(io_file, inter_comm) ) {
    /* inter process communication */
    /* VDCE receive */
#ifdef DEBUG
    fprintf(fd,"vdce send U\n"); 
    fflush(fd);
#endif

    if ( vdce_data_represent(outch2, OUTPUT) ) { /* different data type */

      /* write to the buffer */
      /* write U part */
      send_buf2 = (char *) malloc((matrix_size * matrix_size * 10 + 10) * sizeof(char));
      if (send_buf2 == NULL) {
        fprintf(fd, "malloc error at send_buf2");
      }
      total_size = sizeof(send_buf2);
      bzero(tmp, 10);
      bzero(send_buf2, (matrix_size * matrix_size * 10 + 10));
      sprintf(tmp, "%d ", matrix_size);
      strcat(send_buf2, tmp);
      for ( i = 0 ; i < matrix_size ; i++ ) {
	for ( j = 0 ; j < matrix_size ; j++ ) {
	  bzero(tmp, 10);
	  sprintf( tmp, "%5.3f ", U[i][j]);
	  strcat(send_buf2, tmp);
	}
      }
      strcat(send_buf2,"\n");

#ifdef DEBUG
      fprintf(fd, "U :\n \n" );
#endif

    }
    else {
      total_size = sizeof(matrix_size) + sizeof(float) * matrix_size * matrix_size;

#ifdef DEMOLOG
      fprintf(fd, "total size : %d\n", total_size);
      fflush(fd);
#endif

      send_buf2 = (char *)malloc(total_size + 1);

      bcopy(&matrix_size, send_buf2, sizeof(matrix_size));

#ifdef DEMOLOG
      fprintf(fd, "check 1\n");
      fflush(fd);
#endif

      for ( i = 0 ; i < matrix_size ; i++ ) {
	bcopy(U[i], send_buf2 + sizeof(matrix_size) + i * sizeof(float) * matrix_size, sizeof(float) * matrix_size);
      }

#ifdef DEMOLOG
      fprintf(fd, "check 2\n");
      fflush(fd);
#endif

#ifdef DEBUG
      fprintf(fd,"\nU Matrix\n");
      for ( i = 0 ; i < matrix_size ; i++ ) {
	for ( j = 0 ; j < matrix_size ; j++ ) {
	  fprintf(fd,"%5.3f ", U[i][j]);
	}
	fprintf(fd,"\n");
      }
      fflush(fd);
#endif
    }

    vdce_send(outch2, send_buf2, total_size);
  }
  else {
    /* external file */
    fd1 = fopen(io_file, "w");
 
    if ( fd1 == NULL ) {
      fprintf(fd, "\ncannot find file!\n");
    }
    else {
      fprintf(fd1, "%d\n", matrix_size);
      for ( i = 0 ; i < matrix_size ; i++ ) {
	for ( j = 0 ; j < matrix_size ; j++ ) {
	  fprintf( fd1, "%7.3f", U[i][j]);
	}
	fprintf(fd1, "\n");
      }
      fclose(fd1);
    }
  }

#ifdef DEBUG
  fclose(fd);
#endif
#ifdef DEMOLOG
  fclose(fd);
#endif

/*send CVM execution time */

  total_elapsed = timer_elapsed();
#ifdef DEBUG
  fd = fopen(time_file, "a+");
  fprintf( fd, "%s%d  %d\n", task_name, task_id, total_elapsed);
  fclose(fd);
#endif
  report_execution_time(2,total_elapsed);

}

vdce_def(mytaskinfo)
     VDCE_TASKINFO	*mytaskinfo;
{

  vdce_set_taskname(mytaskinfo,"LU");
	
  vdce_set_inputs(mytaskinfo,"A");

  vdce_set_outputs(mytaskinfo,"L");
  vdce_set_outputs(mytaskinfo,"U");
}
