/**************************************************************************   
 *                                                                        *
 *    File Name : mult.c                                                  *
 *                                                                        *
 *    Author    : KIM, DONGMIN (kimd@top.cis.syr.edu)                     *
 *                                                                        *
 *    Function  : This program is a master program for matrix and vector  *
 *                multiplication for VDCE application.                    *
 *                This program read data from buffer which stores         *
 *                informations received by vdce_recv.                     *
 *                After the computation, it stores L and U matrix to      *
 *                buffer so that they can be sent by vdce_send.           *
 *                                                                        *
 **************************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <sys/types.h>
#include <unistd.h>
#include <math.h>
#include <string.h>

#include "p4.h"

#include "vdce.h"

#define DATA 1

#define inter_comm "??"

#ifdef DEMOLOG
#define time_file "/merlin2/users/demo/vdce/ver1/log/time.dat"
#else
#define time_file "/tmp/time.dat"
#endif

extern char *vdce_get_my_name();
extern	char	*vdce_input_name();
extern	char	*vdce_output_name();
extern	int	vdce_data_represent();

vdce_run(argc,argv)
     int  argc;
     char *argv[];
{
  int nsize;
  int inch1, inch2, outch1;
  char	**buf1;
/**
  char  **buf2;
**/
  char tmp[10];
  char *temp;
  int i, j, k;
  int matrix_size, type, size;
  int vector_size;
  int nslave, my_id, to, row, load, from;
  float **Matrix, **submatrix;
  float *result;
  float *Vector;
  float *incoming;
  float element, zero;
  char	*out1_msg;

  long  total_elapsed;
  FILE  *fd, *fd1;
  char  *task_name;
  int   task_id;
  char  file_name[100];
  char  *io_file;
  int   *counter;
  char  *buf2;
  int   *tmp_matrix_size;
  int   total_size;
  char  *send_buf;
  float *tmp_float;

/**/
  timer_reset();

  task_name = (char *) malloc(10*sizeof(char ));
  if (task_name == NULL) {
    fprintf(fd, "malloc error at task_name");
    fflush(fd);
  }
 
  task_name = vdce_get_my_name();
  task_id = vdce_get_my_id();

#ifdef DEMOLOG
  sprintf(file_name, "/merlin2/users/demo/vdce/ver1/log/mult%d.inter", task_id);
#else
  sprintf(file_name, "/tmp/mult%d.inter", task_id);
#endif

  fd = fopen(file_name, "w");

  p4_initenv(&argc, argv);
#ifdef DEBUG
  fprintf(fd, "after p4_initenv\n");
  fflush(fd);
#endif
  my_id = p4_get_my_id();
  if(my_id != 0){
    slave(); 
    p4_wait_for_end();
    exit(0); 
  }          
/**/
  p4_create_procgroup();
#ifdef DEBUG
  fprintf(fd, "after create_procgroup\n");
  fflush(fd);
#endif

  nslave = p4_num_total_ids() - 1;
  /* printf("my_id :%d  total proc : %d\n", my_id, nslave); */
#ifdef DEBUG
  fprintf(fd, "my_id :%d  total proc : %d\n", my_id, nslave);
  fflush(fd);
#endif

  inch1 = atoi(argv[1]);
  inch2 = atoi(argv[2]);
  outch1 = atoi(argv[3]);

/**
  fprintf(stdout,"in_port1 = %d, in_port2 = %d, out_port1 = %d\n",
	  inch1,inch2,outch1);
**/

#ifdef DEBUG
  fprintf(fd,"in_port1 = %d, in_port2 = %d, out_port1 = %d\n",
	  inch1,inch2,outch1);
  fflush(fd);
#endif

  io_file = (char *) malloc(100*sizeof(char ));
  bzero(io_file, 100);

  io_file = vdce_input_name(task_id, inch1);
  if (io_file == NULL) {
    fprintf(fd, "malloc error at io_file");
    fflush(fd);
  }
#ifdef DEBUG
  fprintf(fd,"io file1 name : %s\n", io_file); 
  fflush(fd);
#endif

  if ( !strcmp(io_file, inter_comm) ) {
    /* inter process communication */
    /* VDCE receive */
#ifdef DEBUG
    fprintf(fd,"vdce receive\n"); 
    fflush(fd);
#endif

    nsize=vdce_recv(inch1, &buf2);

    if ( vdce_data_represent(inch1, INPUT) ) { /* different data type */

      temp = strtok((char * ) buf2, " ");
      matrix_size = atoi(temp);
      Matrix =  (float **) malloc(matrix_size * sizeof(float *));
      if (Matrix == NULL) {
	fprintf(fd, "malloc error at Matrix");
      }
      for ( i = 0 ; i < matrix_size ; i++ ) {
	Matrix[i] = (float *) malloc(matrix_size * sizeof(float));
	if (Matrix[i] == NULL) {
	  fprintf(fd, "malloc error at Matrix[%d]", i);
	}
      }
      for ( i = 1 ; i <= matrix_size ; i++ ) {
	temp = strtok( 0, " ");
	Matrix[i-1][0] = (float)atof(temp);
	for ( j = 1 ; j < matrix_size ; j++ ) {
	  temp = strtok( 0, " ");
	  Matrix[i-1][j] = (float)atof(temp);
	}
      }
    }
    else { /* same data type */

      tmp_matrix_size = (int *)buf2;

      matrix_size = *tmp_matrix_size;

#ifdef DEMOLOG
      fprintf(fd, "receivec size : %d   Matrix size : %d\n", nsize, matrix_size);
      fflush(fd);
#endif

      Matrix =  (float **) malloc(matrix_size * sizeof(float *));
      if (Matrix == NULL) {
	fprintf(fd, "malloc error at Matrix");
      }

      for ( i = 0 ; i < matrix_size ; i++ ) {
	Matrix[i] = (float *)(buf2 + sizeof(int) + i * sizeof(float) * matrix_size);
      }

      if (nsize <= 0 ) {
#ifdef DEBUG
	fprintf(fd,"data has not been arrived\n"); 
	fflush(fd);
#endif
	exit(1);
      }
    }
  }
  else {
    /* external file */
    /* get input data from file */
#ifdef DEBUG
    fprintf(fd,"external file\n"); 
    fflush(fd);
#endif
    fd1 = fopen(io_file, "r");
 
    if ( fd1 == NULL ) {
      printf("\ncannot find file!\n");
      exit(1);
    }
 
    /* read the size of the matrix. */
    fscanf( fd1, "%d", &matrix_size );

    Matrix =  (float **) malloc(matrix_size * sizeof(float *));
    if (Matrix == NULL) {
      fprintf(fd, "malloc error at Matrix");
    }
    for ( i = 0 ; i < matrix_size ; i++ ) {
      Matrix[i] = (float *) malloc(matrix_size * sizeof(float));
      if (Matrix[i] == NULL) {
        fprintf(fd, "malloc error at Matrix[%d]", i);
      }
    }

    /* read from the file */
    for ( i = 0 ; i < matrix_size ; i++ ) {
      for ( j = 0 ; j < matrix_size ; j++ ) {
	fscanf(fd1, "%f", &element);
	Matrix[i][j] = element;
      }
    }
    fclose(fd1);
  }

#ifdef DEBUG
    fprintf(fd, "Original Matrix : \n");
    for ( i = 0 ; i < matrix_size ; i++ ) {
      for ( j = 0 ; j < matrix_size ; j++ ) {
	fprintf(fd, "%f ", Matrix[i][j]);
	fflush(fd);
      }
      fprintf(fd, "\n");
    }

    fprintf(fd, "receivd matrix size :%d", matrix_size);
    fflush(fd);
#endif

  bzero(io_file, 100);

  io_file = vdce_input_name(task_id, inch2);

#ifdef DEBUG
  fprintf(fd,"io file2 name : %s\n", io_file); 
  fflush(fd);
#endif

  if ( !strcmp(io_file, inter_comm) ) {
    /* inter process communication */
    /* VDCE receive */
#ifdef DEBUG
    fprintf(fd,"vdce receive\n"); 
    fflush(fd);
#endif

    nsize=vdce_recv(inch2, &buf2);

    if ( vdce_data_represent(inch2, INPUT) ) { /* different data type */

      temp = strtok((char * ) buf2, " ");
      vector_size = atoi(temp);
      Vector =  (float *) malloc(vector_size * sizeof(float));
      if (Vector == NULL) {
	fprintf(fd, "malloc error at Vector");
      }
      for ( i = 1 ; i <= matrix_size ; i++ ) {
	temp = strtok( 0, " ");
	Vector[i-1] = (float)atof(temp);
      }
    }
    else { /* same data type */

      tmp_matrix_size = (int *)buf2;

      vector_size = *tmp_matrix_size;

#ifdef DEMOLOG
      fprintf(fd, "receivec size : %d   vector size : %d\n", nsize, vector_size);
      fflush(fd);
#endif

      Vector = (float *)(buf2 + sizeof(int));

      if (nsize <= 0 ) {
#ifdef DEBUG
	fprintf(fd,"data has not been arrived\n"); 
	fflush(fd);
#endif
	exit(1);
      }
    }
  }
  else {
    /* external file */
    /* get input data from file */
#ifdef DEBUG
    fprintf(fd,"external file\n"); 
    fflush(fd);
#endif
    fd1 = fopen(io_file, "r");
 
    if ( fd1 == NULL ) {
      printf("\ncannot find file!\n");
      exit(1);
    }
 
    /* read the size of the matrix. */
    fscanf( fd1, "%d", &vector_size );

    Vector =  (float *) malloc(matrix_size * sizeof(float ));
    if (Vector == NULL) {
      fprintf(fd, "malloc error at Vector");
    }

    /* read from the file */
    for ( i = 0 ; i < vector_size ; i++ ) {
      fscanf(fd1, "%f", &element);
      Vector[i] = element;
    }
    fclose(fd1);
  }

#ifdef DEBUG
    fprintf(fd, "Original Vector : \n");
    for ( i = 0 ; i < matrix_size ; i++ ) {
      fprintf(fd, "%f ", Vector[i]);
      fflush(fd);
    }
      fprintf(fd, "\n");

    fprintf(fd, "receivd vector size :%d", vector_size);
    fflush(fd);
#endif

  result =  (float *) malloc(vector_size * sizeof(float ));
  if (result == NULL) {
	fprintf(fd, "malloc error at result");
  }

  load = matrix_size / nslave;

  /* distribute data */
/**/
  for ( to = 1 ; to <= nslave ; to++ ) {
    p4_send(DATA, to, &matrix_size, sizeof(int));
    p4_send(DATA, to, &load, sizeof(int));
    p4_send(DATA, to, Vector, matrix_size * sizeof(float));

    for ( i = 0 ; i < load ; i++ ) {
      p4_send(DATA, to, Matrix[(to - 1) * load + i], matrix_size * sizeof(float));
    }
  }
/**/

  /* collect the result */
  size	= load * sizeof(float);
  type    = DATA;
  for ( from = 1 ; from <= nslave ; from++ ) {
    incoming	= NULL;
    p4_recv(&type, &from, &incoming, &size);
    for ( j = 0 ; j < load ; j++ ) {
      result[ (from - 1) * load + j ] = incoming[j];
    }
    p4_msg_free(incoming);
  }

/*
  printf("\nResult vector\n");
  for ( i = 0 ; i < matrix_size ; i++ ) {
    printf("%5.3f ", result[i]);
  }
  printf("\n");
*/

#ifdef DEBUG
  fprintf(fd, "\nResult vector\n");
  for ( i = 0 ; i < matrix_size ; i++ ) {
    fprintf(fd, "%5.3f ", result[i]);
  }
  fprintf(fd, "\n");
  fflush(fd);
#endif


  bzero(io_file, 100);
  io_file = vdce_output_name(task_id, outch1);
#ifdef DEBUG
  fprintf(fd,"io file name : %s\n", io_file); 
  fflush(fd);
#endif

  if ( !strcmp(io_file, inter_comm) ) {
    /* inter process communication */
    /* VDCE receive */

    if ( vdce_data_represent(outch1, OUTPUT) ) { /* different data type */

      /* write to the buffer */
      send_buf = (char *) malloc((matrix_size * matrix_size * 10 + 10) * sizeof(char));
      if (send_buf == NULL) {
        fprintf(fd, "malloc error at send_buf");
      }
      total_size = sizeof(send_buf);
      bzero(tmp, 10);
      bzero(send_buf, (matrix_size * 10 + 10));
      sprintf(tmp, "%d ", matrix_size);
      strcat(send_buf, tmp);

      for ( i = 0 ; i < matrix_size ; i++ ) {
	bzero(tmp, 10);
	sprintf( tmp, "%5.3f ", result[i]);
	strcat(send_buf, tmp);
      }
      strcat(send_buf,"\n");
    }
    else {

      total_size = sizeof(matrix_size) + sizeof(float) * matrix_size;
      
#ifdef DEMOLOG
      fprintf(fd, "total size : %d\n", total_size);
      fflush(fd);
#endif

      send_buf = (char *)malloc(total_size + 1);

      bcopy(&matrix_size, send_buf, sizeof(matrix_size));

#ifdef DEMOLOG
      fprintf(fd, "check 1\n");
      fflush(fd);
#endif

      for ( i = 0 ; i < matrix_size ; i++ ) {
	bcopy(&result[i], send_buf + sizeof(matrix_size) + i * sizeof(float), sizeof(float));
      }

#ifdef DEMOLOG
      fprintf(fd, "check 2\n");
      fflush(fd);
#endif
    }

    vdce_send(outch1, send_buf, total_size);
  }
  else {
    /* external file */
    fd1 = fopen(io_file, "w");
 
    if ( fd1 == NULL ) {
      fprintf(fd, "\ncannot find file!\n");
    }
    else {
      fprintf(fd1, "%d\n", matrix_size);
      for ( i = 0 ; i < matrix_size ; i++ ) {
	fprintf( fd1, "%7.3f", result[i]);
      }
      fprintf(fd1, "\n");
      fclose(fd1);
    }
  }

  free(Matrix);
  free(Vector);
  free(result);

#ifdef DEBUG
  fclose(fd);
#endif
#ifdef DEMOLOG
  fclose(fd);
#endif

  p4_wait_for_end();
  total_elapsed = timer_elapsed();
#ifdef DEBUG
  fd = fopen(time_file, "a+");
  fprintf( fd, "%s%d  %d\n", task_name, task_id, total_elapsed);
  fclose(fd);
#endif
/*send  CVM execution time */
  report_execution_time(2,total_elapsed); 
  
}

vdce_def(mytaskinfo)
     VDCE_TASKINFO	*mytaskinfo;
{

  vdce_set_taskname(mytaskinfo,"MULT");
	
  vdce_set_inputs(mytaskinfo,"matrix");
  vdce_set_inputs(mytaskinfo,"vector");

  vdce_set_outputs(mytaskinfo,"temp");
}
