/**************************************************************************   
 *									  *
 *    File Name : X_slave.c	 				          *
 *									  *
 *    Author    : KIM, DONGMIN (kimd@top.cis.syr.edu)			  *
 * 									  *
 *    Function  : This slave program takes a NxN matrix A  and            *
 *                Nx1 vectorx B from data file "data_matrix"              *
 *                and computes Nx1 matrix C.		                  *
 *									  *
 **************************************************************************/

#include "p4.h"
/*
#include <stdlib.h>
*/

#define DATA  1
#define N     4
#define MAX   4
 
/*-------------------------------------------------------------------*/

slave() {
  int    i, j, type, to, from, size, my_id;
  int    load, matrix_size, *incoming1;
  float **A, *B, *result, *incoming2;

  my_id = p4_get_my_id();

/*
  printf("===================================\n");
  printf("Slave : %d\n", my_id);
*/

  /* receive the data from master */
  type = DATA;
  size = sizeof(int);
  from = 0;  /* master process */
  incoming1 = NULL;
  p4_recv(&type, &from, &incoming1, &size);
  matrix_size = *incoming1;
  p4_msg_free(incoming1);

  incoming1 = NULL;
  p4_recv(&type, &from, &incoming1, &size);
  load = *incoming1;
  p4_msg_free(incoming1);

  /* set the array A, B, and result using the matrix size */
  A = (float **) calloc(load, sizeof(float *));
  for ( i = 0 ; i < load ; i++ )
    A[i] = (float *) calloc(matrix_size, sizeof(float));

  B = (float *) calloc(matrix_size, sizeof(float));

  result = (float *) calloc(load, sizeof(float));

  /* get the vector B from the master */
  size = matrix_size * sizeof(float);
  incoming2 = NULL;
  p4_recv(&type, &from, &incoming2, &size);
  for ( i = 0 ; i < matrix_size ; i++ ) {
    B[i] = incoming2[i];
  }
  p4_msg_free(incoming2);

  /* get the matrix A from the master */
  for ( i = 0 ; i < load ; i++ ) {
    incoming2 = NULL;
    p4_recv(&type, &from, &incoming2, &size);
    for ( j = 0 ; j < matrix_size ; j++ ) {
      A[i][j] = incoming2[j];
    }
    p4_msg_free(incoming2);
  }

/*
  printf("\n slave %d: Received A\n",my_id);
  for ( j = 0 ; j < load ; j++ ) {
    for ( i = 0 ; i < matrix_size ; i++ ) {
      printf(" %5.2f ", A[j][i]);
    }
    printf("\n");
  }
  printf("\n\n slave %d: Received B\n",my_id);
  for ( i = 0 ; i < matrix_size ; i++ ) {
    printf(" %5.2f ", B[i]);
  }
  printf("\n===================================\n");
*/

  /* compute the allocated data */
  for ( i = 0 ; i < load ; i++ ) {
    result[i] = 0.;
    for ( j = 0 ; j < matrix_size ; j++ ) {
      result[i] = result[i] + A[i][j] * B[j];
    }
  }

/*
  printf("\n\n slave %d: result\n",my_id);
  for ( i = 0 ; i < load ; i++ ) {
    printf(" %5.2f ", result[i]);
  }
  printf("\n");
*/

  to = 0;   /* master */
  size = load * sizeof(float);
  p4_send(DATA, to, result, size);
  free(A);
  free(B);
  free(result);
  

}

/*============================================================*
  Procedure Name : 
  Function       : 
 *============================================================*/
/*
main (argc, argv)
     int     argc;
     char    **argv;
{
  p4_initenv(&argc, argv);
  p4_create_procgroup();
  slave();
  p4_wait_for_end();
  exit(0);
} */
 /* end of slave */


