/**************************************************************************   
 *                                                                        *
 *    File Name : mult.c                                                  *
 *                                                                        *
 *    Author    : KIM, DONGMIN (kimd@top.cis.syr.edu)                     *
 *                                                                        *
 *    Function  : This program is a master program for matrix and vector  *
 *                multiplication for VDCE application.                    *
 *                This program read data from buffer which stores         *
 *                informations received by vdce_recv.                     *
 *                After the computation, it stores L and U matrix to      *
 *                buffer so that they can be sent by vdce_send.           *
 *                                                                        *
 **************************************************************************/




#include <stdio.h>
#include <malloc.h>
#include <sys/types.h>
#include <unistd.h>
#include <math.h>
#include <string.h>

#include "vdce.h"

#define inter_comm "??"

extern char *vdce_get_my_name();
extern	char	*vdce_input_name();
extern	char	*vdce_output_name();
extern	int	vdce_data_represent();

vdce_run(argc,argv)
     int  argc;
     char *argv[];
{
  int nsize;
  int inch1, inch2, outch1;
  char tmp[10];
  char *temp;
  int i, j, k;
  int vector_size;
  int matrix_size;
  float **Matrix, **submatrix;
  float *result;
  float *Vector;
  float element, zero;

  long  total_elapsed;
  FILE  *fd, *fd1;
  char  *task_name;
  int   task_id;
  char  file_name[100];
  char	*out1_msg;
  char  *io_file;
  int   *counter;
  char  *buf2;
  int   *tmp_matrix_size;
  int   total_size;
  char  *send_buf;
  float *tmp_float;

  timer_reset();

  task_name = (char *) malloc(100*sizeof(char ));
  task_name = vdce_get_my_name();
  task_id = vdce_get_my_id();
  sprintf(file_name, "/tmp/%d_%s", task_id, task_name);
 
  fd = fopen(file_name, "w");

  inch1 = atoi(argv[1]);
  inch2 = atoi(argv[2]);
  outch1 = atoi(argv[3]);

  fprintf(fd,"in_port1 = %d, in_port2 = %d, out_port1 = %d\n",
	  inch1,inch2,outch1);
  fflush(fd);

  io_file = (char *) malloc(100*sizeof(char ));
  bzero(io_file, 100);

  io_file = vdce_input_name(task_id, inch1);
  if (io_file == NULL) {
    fprintf(fd, "malloc error at io_file");
    fflush(fd);
  }

  fprintf(fd,"io file1 name : %s\n", io_file); 
  fflush(fd);

  if ( !strcmp(io_file, inter_comm) ) {
    /* inter process communication */
    /* VDCE receive */

    fprintf(fd,"vdce receive\n"); 
    fflush(fd);

    nsize=vdce_recv(inch1, &buf2);

    if ( vdce_data_represent(inch1, INPUT) ) { /* different data type */

      temp = strtok((char * ) buf2, " ");
      matrix_size = atoi(temp);
      Matrix =  (float **) malloc(matrix_size * sizeof(float *));
      if (Matrix == NULL) {
	fprintf(fd, "malloc error at Matrix");
      }
      for ( i = 0 ; i < matrix_size ; i++ ) {
	Matrix[i] = (float *) malloc(matrix_size * sizeof(float));
	if (Matrix[i] == NULL) {
	  fprintf(fd, "malloc error at Matrix[%d]", i);
	}
      }
      for ( i = 1 ; i <= matrix_size ; i++ ) {
	temp = strtok( 0, " ");
	Matrix[i-1][0] = (float)atof(temp);
	for ( j = 1 ; j < matrix_size ; j++ ) {
	  temp = strtok( 0, " ");
	  Matrix[i-1][j] = (float)atof(temp);
	}
      }
    }
    else { /* same data type */

      tmp_matrix_size = (int *)buf2;

      matrix_size = *tmp_matrix_size;

      fprintf(fd, "receivec size : %d   Matrix size : %d\n", nsize, matrix_size);
      fflush(fd);

      Matrix =  (float **) malloc(matrix_size * sizeof(float *));
      if (Matrix == NULL) {
	fprintf(fd, "malloc error at Matrix");
      }

      for ( i = 0 ; i < matrix_size ; i++ ) {
	Matrix[i] = (float *)(buf2 + sizeof(int) + i * sizeof(float) * matrix_size);
      }

      if (nsize <= 0 ) {
	fprintf(fd,"data has not been arrived\n"); 
	fflush(fd);
	exit(1);
      }
    }
  }
  else {
    /* external file */
    /* get input data from file */

    fprintf(fd,"external file\n"); 
    fflush(fd);

    fd1 = fopen(io_file, "r");
 
    if ( fd1 == NULL ) {
      printf("\ncannot find file!\n");
      exit(1);
    }
 
    /* read the size of the matrix. */
    fscanf( fd1, "%d", &matrix_size );

    Matrix =  (float **) malloc(matrix_size * sizeof(float *));
    if (Matrix == NULL) {
      fprintf(fd, "malloc error at Matrix");
    }
    for ( i = 0 ; i < matrix_size ; i++ ) {
      Matrix[i] = (float *) malloc(matrix_size * sizeof(float));
      if (Matrix[i] == NULL) {
        fprintf(fd, "malloc error at Matrix[%d]", i);
      }
    }

    /* read from the file */
    for ( i = 0 ; i < matrix_size ; i++ ) {
      for ( j = 0 ; j < matrix_size ; j++ ) {
	fscanf(fd1, "%f", &element);
	Matrix[i][j] = element;
      }
    }
    fclose(fd1);
  }

  bzero(io_file, 100);

  io_file = vdce_input_name(task_id, inch2);

  fprintf(fd,"io file2 name : %s\n", io_file); 
  fflush(fd);

  if ( !strcmp(io_file, inter_comm) ) {
    /* inter process communication */
    /* VDCE receive */

    fprintf(fd,"vdce receive\n"); 
    fflush(fd);

    nsize=vdce_recv(inch2, &buf2);

    if ( vdce_data_represent(inch2, INPUT) ) { /* different data type */

      temp = strtok((char * ) buf2, " ");
      vector_size = atoi(temp);
      Vector =  (float *) malloc(vector_size * sizeof(float));
      if (Vector == NULL) {
	fprintf(fd, "malloc error at Vector");
      }
      for ( i = 1 ; i <= matrix_size ; i++ ) {
	temp = strtok( 0, " ");
	Vector[i-1] = (float)atof(temp);
      }
    }
    else { /* same data type */

      tmp_matrix_size = (int *)buf2;

      vector_size = *tmp_matrix_size;

      fprintf(fd, "receivec size : %d   vector size : %d\n", nsize, vector_size);
      fflush(fd);

      Vector = (float *)(buf2 + sizeof(int));

      if (nsize <= 0 ) {
	fprintf(fd,"data has not been arrived\n"); 
	fflush(fd);
	exit(1);
      }
    }
  }
  else {
    /* external file */
    /* get input data from file */

    fprintf(fd,"external file\n"); 
    fflush(fd);

    fd1 = fopen(io_file, "r");
 
    if ( fd1 == NULL ) {
      printf("\ncannot find file!\n");
      exit(1);
    }
 
    /* read the size of the matrix. */
    fscanf( fd1, "%d", &vector_size );

    Vector =  (float *) malloc(matrix_size * sizeof(float ));
    if (Vector == NULL) {
      fprintf(fd, "malloc error at Vector");
    }

    /* read from the file */
    for ( i = 0 ; i < vector_size ; i++ ) {
      fscanf(fd1, "%f", &element);
      Vector[i] = element;
    }
    fclose(fd1);
  }

  result =  (float *) malloc(vector_size * sizeof(float ));
  if (result == NULL) {
	fprintf(fd, "malloc error at result");
  }

  compute_mult(Matrix, Vector, result, matrix_size);

  fprintf(fd, "\nResult vector\n");
  fflush(fd);

  bzero(io_file, 100);
  io_file = vdce_output_name(task_id, outch1);

  fprintf(fd,"io file name : %s\n", io_file); 
  fflush(fd);

  if ( !strcmp(io_file, inter_comm) ) {
    /* inter process communication */
    /* VDCE receive */

    if ( vdce_data_represent(outch1, OUTPUT) ) { /* different data type */

      /* write to the buffer */
      send_buf = (char *) malloc((matrix_size * matrix_size * 10 + 10) * sizeof(char));
      if (send_buf == NULL) {
        fprintf(fd, "malloc error at send_buf");
      }
      total_size = sizeof(send_buf);
      bzero(tmp, 10);
      bzero(send_buf, (matrix_size * 10 + 10));
      sprintf(tmp, "%d ", matrix_size);
      strcat(send_buf, tmp);

      for ( i = 0 ; i < matrix_size ; i++ ) {
	bzero(tmp, 10);
	sprintf( tmp, "%5.3f ", result[i]);
	strcat(send_buf, tmp);
      }
      strcat(send_buf,"\n");
    }
    else {

      total_size = sizeof(matrix_size) + sizeof(float) * matrix_size;
      
      fprintf(fd, "total size : %d\n", total_size);
      fflush(fd);

      send_buf = (char *)malloc(total_size + 1);

      bcopy(&matrix_size, send_buf, sizeof(matrix_size));

      for ( i = 0 ; i < matrix_size ; i++ ) {
	bcopy(&result[i], send_buf + sizeof(matrix_size) + i * sizeof(float), sizeof(float));
      }
    }

    vdce_send(outch1, send_buf, total_size);
  }
  else {
    /* external file */
    fd1 = fopen(io_file, "w");
 
    if ( fd1 == NULL ) {
      fprintf(fd, "\ncannot find file!\n");
    }
    else {
      fprintf(fd1, "%d\n", matrix_size);
      for ( i = 0 ; i < matrix_size ; i++ ) {
	fprintf( fd1, "%7.3f", result[i]);
      }
      fprintf(fd1, "\n");
      fclose(fd1);
    }
  }

  total_elapsed = timer_elapsed();

  task_name = vdce_get_my_name();
  task_id = vdce_get_my_id();

  fprintf(fd,"%s%d  %d\n", task_name, task_id, total_elapsed);
  fclose(fd);

  report_execution_time(2,total_elapsed);

}

vdce_def(mytaskinfo)
     VDCE_TASKINFO	*mytaskinfo;
{

  vdce_set_taskname(mytaskinfo,"MULT");
	
  vdce_set_inputs(mytaskinfo,"matrix");
  vdce_set_inputs(mytaskinfo,"vector");

  vdce_set_outputs(mytaskinfo,"temp");
}

compute_mult(A, B, result, matrix_size)
float **A, *B, *result;
int   matrix_size;

{
  int    i, j;


  /* compute the allocated data */
  for ( i = 0 ; i < matrix_size ; i++ ) {
    result[i] = 0.;
    for ( j = 0 ; j < matrix_size ; j++ ) {
      result[i] = result[i] + A[i][j] * B[j];
    }
  }

/*
  printf("\n\n slave %d: result\n",my_id);
  for ( i = 0 ; i < load ; i++ ) {
    printf(" %5.2f ", result[i]);
  }
  printf("\n");
*/

}

